from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    ProcessState,
    ProcessType,
    optional_field,
)
from lqs.interface.core import models


class Hook(CommonModel):
    workflow_id: UUID = Field(
        ..., description="The ID of the workflow to which the hook belongs."
    )
    trigger_process: ProcessType = Field(
        ..., description="The process type that triggers the hook."
    )
    trigger_state: ProcessState = Field(
        ..., description="The process state that triggers the hook."
    )
    name: Optional[str] = Field(
        None, description="The name of the hook for reference (not unique)."
    )
    note: Optional[str] = Field(
        None, description="A general note about the hook for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the hook."
    )
    managed: bool = Field(..., description="Whether the hook is managed by the system.")
    disabled: bool = Field(
        ..., description="Whether the hook is disabled and cannot be triggered."
    )
    uri: Optional[str] = Field(
        None, description="The URI to which the hook should send requests."
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            ["name", "note", "context", "uri"]
        )
    )

    def fetch_workflow(self) -> "models.Workflow":
        return self._app.fetch.workflow(self.workflow_id).data


class HookDataResponse(DataResponseModel[Hook]):
    pass


class HookListResponse(PaginationModel[Hook]):
    pass


class HookForWorkflowCreateRequest(BaseModel):
    trigger_process: ProcessType
    trigger_state: ProcessState
    name: Optional[str] = None
    note: Optional[str] = None
    context: Optional[dict] = None
    disabled: bool = False
    managed: bool = False
    uri: Optional[str] = None
    secret: Optional[str] = None


class HookCreateRequest(HookForWorkflowCreateRequest):
    workflow_id: UUID


class HookUpdateRequest(BaseModel):
    trigger_process: ProcessType = optional_field
    trigger_state: ProcessState = optional_field
    name: Optional[str] = optional_field
    note: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    disabled: bool = optional_field
    uri: Optional[str] = optional_field
    secret: Optional[str] = optional_field



# Generated from RosMessageParser.g4 by ANTLR 4.13.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    return [
        4,
        1,
        45,
        193,
        2,
        0,
        7,
        0,
        2,
        1,
        7,
        1,
        2,
        2,
        7,
        2,
        2,
        3,
        7,
        3,
        2,
        4,
        7,
        4,
        2,
        5,
        7,
        5,
        2,
        6,
        7,
        6,
        2,
        7,
        7,
        7,
        2,
        8,
        7,
        8,
        2,
        9,
        7,
        9,
        2,
        10,
        7,
        10,
        2,
        11,
        7,
        11,
        2,
        12,
        7,
        12,
        2,
        13,
        7,
        13,
        2,
        14,
        7,
        14,
        2,
        15,
        7,
        15,
        2,
        16,
        7,
        16,
        2,
        17,
        7,
        17,
        2,
        18,
        7,
        18,
        2,
        19,
        7,
        19,
        2,
        20,
        7,
        20,
        2,
        21,
        7,
        21,
        2,
        22,
        7,
        22,
        2,
        23,
        7,
        23,
        2,
        24,
        7,
        24,
        2,
        25,
        7,
        25,
        1,
        0,
        1,
        0,
        1,
        0,
        3,
        0,
        56,
        8,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        3,
        1,
        4,
        1,
        4,
        5,
        4,
        75,
        8,
        4,
        10,
        4,
        12,
        4,
        78,
        9,
        4,
        1,
        4,
        1,
        4,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        6,
        1,
        6,
        1,
        6,
        5,
        6,
        89,
        8,
        6,
        10,
        6,
        12,
        6,
        92,
        9,
        6,
        1,
        7,
        1,
        7,
        3,
        7,
        96,
        8,
        7,
        1,
        7,
        1,
        7,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        3,
        8,
        110,
        8,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        3,
        8,
        117,
        8,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        3,
        8,
        124,
        8,
        8,
        1,
        9,
        1,
        9,
        1,
        9,
        3,
        9,
        129,
        8,
        9,
        1,
        10,
        1,
        10,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        1,
        11,
        3,
        11,
        139,
        8,
        11,
        1,
        12,
        1,
        12,
        1,
        12,
        1,
        12,
        3,
        12,
        145,
        8,
        12,
        1,
        13,
        1,
        13,
        3,
        13,
        149,
        8,
        13,
        1,
        14,
        1,
        14,
        1,
        14,
        1,
        14,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        1,
        15,
        3,
        15,
        165,
        8,
        15,
        1,
        16,
        1,
        16,
        1,
        17,
        1,
        17,
        1,
        18,
        1,
        18,
        1,
        19,
        1,
        19,
        1,
        20,
        1,
        20,
        1,
        21,
        1,
        21,
        1,
        22,
        3,
        22,
        180,
        8,
        22,
        1,
        22,
        1,
        22,
        1,
        23,
        3,
        23,
        185,
        8,
        23,
        1,
        23,
        1,
        23,
        1,
        24,
        1,
        24,
        1,
        25,
        1,
        25,
        1,
        25,
        0,
        0,
        26,
        0,
        2,
        4,
        6,
        8,
        10,
        12,
        14,
        16,
        18,
        20,
        22,
        24,
        26,
        28,
        30,
        32,
        34,
        36,
        38,
        40,
        42,
        44,
        46,
        48,
        50,
        0,
        9,
        2,
        0,
        23,
        23,
        39,
        39,
        3,
        0,
        1,
        16,
        27,
        29,
        34,
        34,
        2,
        0,
        18,
        18,
        35,
        35,
        2,
        0,
        19,
        19,
        36,
        36,
        1,
        0,
        2,
        11,
        1,
        0,
        12,
        13,
        1,
        0,
        15,
        16,
        1,
        0,
        21,
        22,
        1,
        0,
        27,
        28,
        189,
        0,
        55,
        1,
        0,
        0,
        0,
        2,
        57,
        1,
        0,
        0,
        0,
        4,
        60,
        1,
        0,
        0,
        0,
        6,
        67,
        1,
        0,
        0,
        0,
        8,
        72,
        1,
        0,
        0,
        0,
        10,
        81,
        1,
        0,
        0,
        0,
        12,
        90,
        1,
        0,
        0,
        0,
        14,
        95,
        1,
        0,
        0,
        0,
        16,
        123,
        1,
        0,
        0,
        0,
        18,
        128,
        1,
        0,
        0,
        0,
        20,
        130,
        1,
        0,
        0,
        0,
        22,
        138,
        1,
        0,
        0,
        0,
        24,
        144,
        1,
        0,
        0,
        0,
        26,
        148,
        1,
        0,
        0,
        0,
        28,
        150,
        1,
        0,
        0,
        0,
        30,
        164,
        1,
        0,
        0,
        0,
        32,
        166,
        1,
        0,
        0,
        0,
        34,
        168,
        1,
        0,
        0,
        0,
        36,
        170,
        1,
        0,
        0,
        0,
        38,
        172,
        1,
        0,
        0,
        0,
        40,
        174,
        1,
        0,
        0,
        0,
        42,
        176,
        1,
        0,
        0,
        0,
        44,
        179,
        1,
        0,
        0,
        0,
        46,
        184,
        1,
        0,
        0,
        0,
        48,
        188,
        1,
        0,
        0,
        0,
        50,
        190,
        1,
        0,
        0,
        0,
        52,
        56,
        3,
        2,
        1,
        0,
        53,
        56,
        3,
        4,
        2,
        0,
        54,
        56,
        3,
        6,
        3,
        0,
        55,
        52,
        1,
        0,
        0,
        0,
        55,
        53,
        1,
        0,
        0,
        0,
        55,
        54,
        1,
        0,
        0,
        0,
        56,
        1,
        1,
        0,
        0,
        0,
        57,
        58,
        3,
        12,
        6,
        0,
        58,
        59,
        5,
        0,
        0,
        1,
        59,
        3,
        1,
        0,
        0,
        0,
        60,
        61,
        3,
        12,
        6,
        0,
        61,
        62,
        5,
        24,
        0,
        0,
        62,
        63,
        3,
        12,
        6,
        0,
        63,
        64,
        5,
        24,
        0,
        0,
        64,
        65,
        3,
        12,
        6,
        0,
        65,
        66,
        5,
        0,
        0,
        1,
        66,
        5,
        1,
        0,
        0,
        0,
        67,
        68,
        3,
        12,
        6,
        0,
        68,
        69,
        5,
        24,
        0,
        0,
        69,
        70,
        3,
        12,
        6,
        0,
        70,
        71,
        5,
        0,
        0,
        1,
        71,
        7,
        1,
        0,
        0,
        0,
        72,
        76,
        3,
        12,
        6,
        0,
        73,
        75,
        3,
        10,
        5,
        0,
        74,
        73,
        1,
        0,
        0,
        0,
        75,
        78,
        1,
        0,
        0,
        0,
        76,
        74,
        1,
        0,
        0,
        0,
        76,
        77,
        1,
        0,
        0,
        0,
        77,
        79,
        1,
        0,
        0,
        0,
        78,
        76,
        1,
        0,
        0,
        0,
        79,
        80,
        5,
        0,
        0,
        1,
        80,
        9,
        1,
        0,
        0,
        0,
        81,
        82,
        5,
        30,
        0,
        0,
        82,
        83,
        3,
        24,
        12,
        0,
        83,
        84,
        3,
        12,
        6,
        0,
        84,
        11,
        1,
        0,
        0,
        0,
        85,
        89,
        3,
        14,
        7,
        0,
        86,
        89,
        3,
        16,
        8,
        0,
        87,
        89,
        3,
        18,
        9,
        0,
        88,
        85,
        1,
        0,
        0,
        0,
        88,
        86,
        1,
        0,
        0,
        0,
        88,
        87,
        1,
        0,
        0,
        0,
        89,
        92,
        1,
        0,
        0,
        0,
        90,
        88,
        1,
        0,
        0,
        0,
        90,
        91,
        1,
        0,
        0,
        0,
        91,
        13,
        1,
        0,
        0,
        0,
        92,
        90,
        1,
        0,
        0,
        0,
        93,
        96,
        3,
        22,
        11,
        0,
        94,
        96,
        3,
        26,
        13,
        0,
        95,
        93,
        1,
        0,
        0,
        0,
        95,
        94,
        1,
        0,
        0,
        0,
        96,
        97,
        1,
        0,
        0,
        0,
        97,
        98,
        3,
        20,
        10,
        0,
        98,
        15,
        1,
        0,
        0,
        0,
        99,
        100,
        3,
        32,
        16,
        0,
        100,
        101,
        3,
        20,
        10,
        0,
        101,
        102,
        5,
        20,
        0,
        0,
        102,
        103,
        3,
        44,
        22,
        0,
        103,
        124,
        1,
        0,
        0,
        0,
        104,
        105,
        3,
        34,
        17,
        0,
        105,
        106,
        3,
        20,
        10,
        0,
        106,
        109,
        5,
        20,
        0,
        0,
        107,
        110,
        3,
        44,
        22,
        0,
        108,
        110,
        3,
        46,
        23,
        0,
        109,
        107,
        1,
        0,
        0,
        0,
        109,
        108,
        1,
        0,
        0,
        0,
        110,
        124,
        1,
        0,
        0,
        0,
        111,
        112,
        3,
        40,
        20,
        0,
        112,
        113,
        3,
        20,
        10,
        0,
        113,
        116,
        5,
        20,
        0,
        0,
        114,
        117,
        3,
        48,
        24,
        0,
        115,
        117,
        3,
        44,
        22,
        0,
        116,
        114,
        1,
        0,
        0,
        0,
        116,
        115,
        1,
        0,
        0,
        0,
        117,
        124,
        1,
        0,
        0,
        0,
        118,
        119,
        3,
        38,
        19,
        0,
        119,
        120,
        3,
        20,
        10,
        0,
        120,
        121,
        5,
        38,
        0,
        0,
        121,
        122,
        3,
        50,
        25,
        0,
        122,
        124,
        1,
        0,
        0,
        0,
        123,
        99,
        1,
        0,
        0,
        0,
        123,
        104,
        1,
        0,
        0,
        0,
        123,
        111,
        1,
        0,
        0,
        0,
        123,
        118,
        1,
        0,
        0,
        0,
        124,
        17,
        1,
        0,
        0,
        0,
        125,
        126,
        7,
        0,
        0,
        0,
        126,
        129,
        5,
        44,
        0,
        0,
        127,
        129,
        7,
        0,
        0,
        0,
        128,
        125,
        1,
        0,
        0,
        0,
        128,
        127,
        1,
        0,
        0,
        0,
        129,
        19,
        1,
        0,
        0,
        0,
        130,
        131,
        7,
        1,
        0,
        0,
        131,
        21,
        1,
        0,
        0,
        0,
        132,
        139,
        3,
        32,
        16,
        0,
        133,
        139,
        3,
        34,
        17,
        0,
        134,
        139,
        3,
        36,
        18,
        0,
        135,
        139,
        3,
        40,
        20,
        0,
        136,
        139,
        3,
        38,
        19,
        0,
        137,
        139,
        3,
        24,
        12,
        0,
        138,
        132,
        1,
        0,
        0,
        0,
        138,
        133,
        1,
        0,
        0,
        0,
        138,
        134,
        1,
        0,
        0,
        0,
        138,
        135,
        1,
        0,
        0,
        0,
        138,
        136,
        1,
        0,
        0,
        0,
        138,
        137,
        1,
        0,
        0,
        0,
        139,
        23,
        1,
        0,
        0,
        0,
        140,
        145,
        5,
        29,
        0,
        0,
        141,
        142,
        5,
        29,
        0,
        0,
        142,
        143,
        5,
        17,
        0,
        0,
        143,
        145,
        5,
        29,
        0,
        0,
        144,
        140,
        1,
        0,
        0,
        0,
        144,
        141,
        1,
        0,
        0,
        0,
        145,
        25,
        1,
        0,
        0,
        0,
        146,
        149,
        3,
        28,
        14,
        0,
        147,
        149,
        3,
        30,
        15,
        0,
        148,
        146,
        1,
        0,
        0,
        0,
        148,
        147,
        1,
        0,
        0,
        0,
        149,
        27,
        1,
        0,
        0,
        0,
        150,
        151,
        3,
        22,
        11,
        0,
        151,
        152,
        7,
        2,
        0,
        0,
        152,
        153,
        7,
        3,
        0,
        0,
        153,
        29,
        1,
        0,
        0,
        0,
        154,
        155,
        3,
        22,
        11,
        0,
        155,
        156,
        5,
        18,
        0,
        0,
        156,
        157,
        5,
        25,
        0,
        0,
        157,
        158,
        5,
        19,
        0,
        0,
        158,
        165,
        1,
        0,
        0,
        0,
        159,
        160,
        3,
        22,
        11,
        0,
        160,
        161,
        5,
        35,
        0,
        0,
        161,
        162,
        5,
        37,
        0,
        0,
        162,
        163,
        5,
        36,
        0,
        0,
        163,
        165,
        1,
        0,
        0,
        0,
        164,
        154,
        1,
        0,
        0,
        0,
        164,
        159,
        1,
        0,
        0,
        0,
        165,
        31,
        1,
        0,
        0,
        0,
        166,
        167,
        7,
        4,
        0,
        0,
        167,
        33,
        1,
        0,
        0,
        0,
        168,
        169,
        7,
        5,
        0,
        0,
        169,
        35,
        1,
        0,
        0,
        0,
        170,
        171,
        7,
        6,
        0,
        0,
        171,
        37,
        1,
        0,
        0,
        0,
        172,
        173,
        5,
        14,
        0,
        0,
        173,
        39,
        1,
        0,
        0,
        0,
        174,
        175,
        5,
        1,
        0,
        0,
        175,
        41,
        1,
        0,
        0,
        0,
        176,
        177,
        7,
        7,
        0,
        0,
        177,
        43,
        1,
        0,
        0,
        0,
        178,
        180,
        3,
        42,
        21,
        0,
        179,
        178,
        1,
        0,
        0,
        0,
        179,
        180,
        1,
        0,
        0,
        0,
        180,
        181,
        1,
        0,
        0,
        0,
        181,
        182,
        5,
        25,
        0,
        0,
        182,
        45,
        1,
        0,
        0,
        0,
        183,
        185,
        3,
        42,
        21,
        0,
        184,
        183,
        1,
        0,
        0,
        0,
        184,
        185,
        1,
        0,
        0,
        0,
        185,
        186,
        1,
        0,
        0,
        0,
        186,
        187,
        5,
        26,
        0,
        0,
        187,
        47,
        1,
        0,
        0,
        0,
        188,
        189,
        7,
        8,
        0,
        0,
        189,
        49,
        1,
        0,
        0,
        0,
        190,
        191,
        5,
        42,
        0,
        0,
        191,
        51,
        1,
        0,
        0,
        0,
        15,
        55,
        76,
        88,
        90,
        95,
        109,
        116,
        123,
        128,
        138,
        144,
        148,
        164,
        179,
        184,
    ]


class RosMessageParser(Parser):

    grammarFileName = "RosMessageParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "'bool'",
        "'int8'",
        "'uint8'",
        "'byte'",
        "'char'",
        "'int16'",
        "'uint16'",
        "'int32'",
        "'uint32'",
        "'int64'",
        "'uint64'",
        "'float32'",
        "'float64'",
        "'string'",
        "'time'",
        "'duration'",
        "'/'",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "'+'",
        "'-'",
        "<INVALID>",
        "'---'",
        "<INVALID>",
        "<INVALID>",
        "'True'",
        "'False'",
    ]

    symbolicNames = [
        "<INVALID>",
        "BOOL",
        "INT8",
        "UINT8",
        "BYTE",
        "CHAR",
        "INT16",
        "UINT16",
        "INT32",
        "UINT32",
        "INT64",
        "UINT64",
        "FLOAT32",
        "FLOAT64",
        "STRING",
        "TIME",
        "DURATION",
        "SLASH",
        "OPEN_BRACKET",
        "CLOSE_BRACKET",
        "ASSIGNMENT",
        "PLUS",
        "MINUS",
        "HASH",
        "MESSAGE_SEPARATOR",
        "INTEGER_LITERAL",
        "REAL_LITERAL",
        "TRUE",
        "FALSE",
        "IDENTIFIER",
        "ROSBAG_MESSAGE_SEPARATOR",
        "WHITESPACES",
        "NEWLINES",
        "NEWLINE",
        "STRING_IDENTIFIER",
        "STRING_OPEN_BRACKET",
        "STRING_CLOSE_BRACKET",
        "STRING_INTEGER_LITERAL",
        "STRING_ASSIGNMENT",
        "STRING_HASH",
        "STRING_WHITESPACES",
        "STRING_NEWLINE",
        "STRING_VALUE",
        "STRIN_ASSIGNMENT_NEWLINE",
        "COMMENT",
        "COMMENT_NEWLINE",
    ]

    RULE_ros_file_input = 0
    RULE_ros_message_input = 1
    RULE_ros_action_input = 2
    RULE_ros_service_input = 3
    RULE_rosbag_input = 4
    RULE_rosbag_nested_message = 5
    RULE_ros_message = 6
    RULE_field_declaration = 7
    RULE_constant_declaration = 8
    RULE_comment = 9
    RULE_identifier = 10
    RULE_type_ = 11
    RULE_ros_type = 12
    RULE_array_type = 13
    RULE_variable_array_type = 14
    RULE_fixed_array_type = 15
    RULE_integral_type = 16
    RULE_floating_point_type = 17
    RULE_temporal_type = 18
    RULE_string_type = 19
    RULE_boolean_type = 20
    RULE_sign = 21
    RULE_integral_value = 22
    RULE_floating_point_value = 23
    RULE_bool_value = 24
    RULE_string_value = 25

    ruleNames = [
        "ros_file_input",
        "ros_message_input",
        "ros_action_input",
        "ros_service_input",
        "rosbag_input",
        "rosbag_nested_message",
        "ros_message",
        "field_declaration",
        "constant_declaration",
        "comment",
        "identifier",
        "type_",
        "ros_type",
        "array_type",
        "variable_array_type",
        "fixed_array_type",
        "integral_type",
        "floating_point_type",
        "temporal_type",
        "string_type",
        "boolean_type",
        "sign",
        "integral_value",
        "floating_point_value",
        "bool_value",
        "string_value",
    ]

    EOF = Token.EOF
    BOOL = 1
    INT8 = 2
    UINT8 = 3
    BYTE = 4
    CHAR = 5
    INT16 = 6
    UINT16 = 7
    INT32 = 8
    UINT32 = 9
    INT64 = 10
    UINT64 = 11
    FLOAT32 = 12
    FLOAT64 = 13
    STRING = 14
    TIME = 15
    DURATION = 16
    SLASH = 17
    OPEN_BRACKET = 18
    CLOSE_BRACKET = 19
    ASSIGNMENT = 20
    PLUS = 21
    MINUS = 22
    HASH = 23
    MESSAGE_SEPARATOR = 24
    INTEGER_LITERAL = 25
    REAL_LITERAL = 26
    TRUE = 27
    FALSE = 28
    IDENTIFIER = 29
    ROSBAG_MESSAGE_SEPARATOR = 30
    WHITESPACES = 31
    NEWLINES = 32
    NEWLINE = 33
    STRING_IDENTIFIER = 34
    STRING_OPEN_BRACKET = 35
    STRING_CLOSE_BRACKET = 36
    STRING_INTEGER_LITERAL = 37
    STRING_ASSIGNMENT = 38
    STRING_HASH = 39
    STRING_WHITESPACES = 40
    STRING_NEWLINE = 41
    STRING_VALUE = 42
    STRIN_ASSIGNMENT_NEWLINE = 43
    COMMENT = 44
    COMMENT_NEWLINE = 45

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.0")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class Ros_file_inputContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ros_message_input(self):
            return self.getTypedRuleContext(
                RosMessageParser.Ros_message_inputContext, 0
            )

        def ros_action_input(self):
            return self.getTypedRuleContext(RosMessageParser.Ros_action_inputContext, 0)

        def ros_service_input(self):
            return self.getTypedRuleContext(
                RosMessageParser.Ros_service_inputContext, 0
            )

        def getRuleIndex(self):
            return RosMessageParser.RULE_ros_file_input

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRos_file_input"):
                listener.enterRos_file_input(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRos_file_input"):
                listener.exitRos_file_input(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRos_file_input"):
                return visitor.visitRos_file_input(self)
            else:
                return visitor.visitChildren(self)

    def ros_file_input(self):

        localctx = RosMessageParser.Ros_file_inputContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_ros_file_input)
        try:
            self.state = 55
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 0, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 52
                self.ros_message_input()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 53
                self.ros_action_input()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 54
                self.ros_service_input()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ros_message_inputContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ros_message(self):
            return self.getTypedRuleContext(RosMessageParser.Ros_messageContext, 0)

        def EOF(self):
            return self.getToken(RosMessageParser.EOF, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_ros_message_input

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRos_message_input"):
                listener.enterRos_message_input(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRos_message_input"):
                listener.exitRos_message_input(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRos_message_input"):
                return visitor.visitRos_message_input(self)
            else:
                return visitor.visitChildren(self)

    def ros_message_input(self):

        localctx = RosMessageParser.Ros_message_inputContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 2, self.RULE_ros_message_input)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 57
            self.ros_message()
            self.state = 58
            self.match(RosMessageParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ros_action_inputContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ros_message(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(RosMessageParser.Ros_messageContext)
            else:
                return self.getTypedRuleContext(RosMessageParser.Ros_messageContext, i)

        def MESSAGE_SEPARATOR(self, i: int = None):
            if i is None:
                return self.getTokens(RosMessageParser.MESSAGE_SEPARATOR)
            else:
                return self.getToken(RosMessageParser.MESSAGE_SEPARATOR, i)

        def EOF(self):
            return self.getToken(RosMessageParser.EOF, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_ros_action_input

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRos_action_input"):
                listener.enterRos_action_input(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRos_action_input"):
                listener.exitRos_action_input(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRos_action_input"):
                return visitor.visitRos_action_input(self)
            else:
                return visitor.visitChildren(self)

    def ros_action_input(self):

        localctx = RosMessageParser.Ros_action_inputContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_ros_action_input)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 60
            self.ros_message()
            self.state = 61
            self.match(RosMessageParser.MESSAGE_SEPARATOR)
            self.state = 62
            self.ros_message()
            self.state = 63
            self.match(RosMessageParser.MESSAGE_SEPARATOR)
            self.state = 64
            self.ros_message()
            self.state = 65
            self.match(RosMessageParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ros_service_inputContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ros_message(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(RosMessageParser.Ros_messageContext)
            else:
                return self.getTypedRuleContext(RosMessageParser.Ros_messageContext, i)

        def MESSAGE_SEPARATOR(self):
            return self.getToken(RosMessageParser.MESSAGE_SEPARATOR, 0)

        def EOF(self):
            return self.getToken(RosMessageParser.EOF, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_ros_service_input

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRos_service_input"):
                listener.enterRos_service_input(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRos_service_input"):
                listener.exitRos_service_input(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRos_service_input"):
                return visitor.visitRos_service_input(self)
            else:
                return visitor.visitChildren(self)

    def ros_service_input(self):

        localctx = RosMessageParser.Ros_service_inputContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 6, self.RULE_ros_service_input)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 67
            self.ros_message()
            self.state = 68
            self.match(RosMessageParser.MESSAGE_SEPARATOR)
            self.state = 69
            self.ros_message()
            self.state = 70
            self.match(RosMessageParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Rosbag_inputContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ros_message(self):
            return self.getTypedRuleContext(RosMessageParser.Ros_messageContext, 0)

        def EOF(self):
            return self.getToken(RosMessageParser.EOF, 0)

        def rosbag_nested_message(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    RosMessageParser.Rosbag_nested_messageContext
                )
            else:
                return self.getTypedRuleContext(
                    RosMessageParser.Rosbag_nested_messageContext, i
                )

        def getRuleIndex(self):
            return RosMessageParser.RULE_rosbag_input

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRosbag_input"):
                listener.enterRosbag_input(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRosbag_input"):
                listener.exitRosbag_input(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRosbag_input"):
                return visitor.visitRosbag_input(self)
            else:
                return visitor.visitChildren(self)

    def rosbag_input(self):

        localctx = RosMessageParser.Rosbag_inputContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_rosbag_input)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 72
            self.ros_message()
            self.state = 76
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == 30:
                self.state = 73
                self.rosbag_nested_message()
                self.state = 78
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 79
            self.match(RosMessageParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Rosbag_nested_messageContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ROSBAG_MESSAGE_SEPARATOR(self):
            return self.getToken(RosMessageParser.ROSBAG_MESSAGE_SEPARATOR, 0)

        def ros_type(self):
            return self.getTypedRuleContext(RosMessageParser.Ros_typeContext, 0)

        def ros_message(self):
            return self.getTypedRuleContext(RosMessageParser.Ros_messageContext, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_rosbag_nested_message

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRosbag_nested_message"):
                listener.enterRosbag_nested_message(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRosbag_nested_message"):
                listener.exitRosbag_nested_message(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRosbag_nested_message"):
                return visitor.visitRosbag_nested_message(self)
            else:
                return visitor.visitChildren(self)

    def rosbag_nested_message(self):

        localctx = RosMessageParser.Rosbag_nested_messageContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 10, self.RULE_rosbag_nested_message)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 81
            self.match(RosMessageParser.ROSBAG_MESSAGE_SEPARATOR)
            self.state = 82
            self.ros_type()
            self.state = 83
            self.ros_message()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ros_messageContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def field_declaration(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    RosMessageParser.Field_declarationContext
                )
            else:
                return self.getTypedRuleContext(
                    RosMessageParser.Field_declarationContext, i
                )

        def constant_declaration(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(
                    RosMessageParser.Constant_declarationContext
                )
            else:
                return self.getTypedRuleContext(
                    RosMessageParser.Constant_declarationContext, i
                )

        def comment(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(RosMessageParser.CommentContext)
            else:
                return self.getTypedRuleContext(RosMessageParser.CommentContext, i)

        def getRuleIndex(self):
            return RosMessageParser.RULE_ros_message

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRos_message"):
                listener.enterRos_message(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRos_message"):
                listener.exitRos_message(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRos_message"):
                return visitor.visitRos_message(self)
            else:
                return visitor.visitChildren(self)

    def ros_message(self):

        localctx = RosMessageParser.Ros_messageContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_ros_message)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 90
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while ((_la) & ~0x3F) == 0 and ((1 << _la) & 550301204478) != 0:
                self.state = 88
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 2, self._ctx)
                if la_ == 1:
                    self.state = 85
                    self.field_declaration()
                    pass

                elif la_ == 2:
                    self.state = 86
                    self.constant_declaration()
                    pass

                elif la_ == 3:
                    self.state = 87
                    self.comment()
                    pass

                self.state = 92
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Field_declarationContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier(self):
            return self.getTypedRuleContext(RosMessageParser.IdentifierContext, 0)

        def type_(self):
            return self.getTypedRuleContext(RosMessageParser.Type_Context, 0)

        def array_type(self):
            return self.getTypedRuleContext(RosMessageParser.Array_typeContext, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_field_declaration

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterField_declaration"):
                listener.enterField_declaration(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitField_declaration"):
                listener.exitField_declaration(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitField_declaration"):
                return visitor.visitField_declaration(self)
            else:
                return visitor.visitChildren(self)

    def field_declaration(self):

        localctx = RosMessageParser.Field_declarationContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 14, self.RULE_field_declaration)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 95
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 4, self._ctx)
            if la_ == 1:
                self.state = 93
                self.type_()
                pass

            elif la_ == 2:
                self.state = 94
                self.array_type()
                pass

            self.state = 97
            self.identifier()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Constant_declarationContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def integral_type(self):
            return self.getTypedRuleContext(RosMessageParser.Integral_typeContext, 0)

        def identifier(self):
            return self.getTypedRuleContext(RosMessageParser.IdentifierContext, 0)

        def ASSIGNMENT(self):
            return self.getToken(RosMessageParser.ASSIGNMENT, 0)

        def integral_value(self):
            return self.getTypedRuleContext(RosMessageParser.Integral_valueContext, 0)

        def floating_point_type(self):
            return self.getTypedRuleContext(
                RosMessageParser.Floating_point_typeContext, 0
            )

        def floating_point_value(self):
            return self.getTypedRuleContext(
                RosMessageParser.Floating_point_valueContext, 0
            )

        def boolean_type(self):
            return self.getTypedRuleContext(RosMessageParser.Boolean_typeContext, 0)

        def bool_value(self):
            return self.getTypedRuleContext(RosMessageParser.Bool_valueContext, 0)

        def string_type(self):
            return self.getTypedRuleContext(RosMessageParser.String_typeContext, 0)

        def STRING_ASSIGNMENT(self):
            return self.getToken(RosMessageParser.STRING_ASSIGNMENT, 0)

        def string_value(self):
            return self.getTypedRuleContext(RosMessageParser.String_valueContext, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_constant_declaration

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterConstant_declaration"):
                listener.enterConstant_declaration(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitConstant_declaration"):
                listener.exitConstant_declaration(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitConstant_declaration"):
                return visitor.visitConstant_declaration(self)
            else:
                return visitor.visitChildren(self)

    def constant_declaration(self):

        localctx = RosMessageParser.Constant_declarationContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 16, self.RULE_constant_declaration)
        try:
            self.state = 123
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [2, 3, 4, 5, 6, 7, 8, 9, 10, 11]:
                self.enterOuterAlt(localctx, 1)
                self.state = 99
                self.integral_type()
                self.state = 100
                self.identifier()
                self.state = 101
                self.match(RosMessageParser.ASSIGNMENT)
                self.state = 102
                self.integral_value()
                pass
            elif token in [12, 13]:
                self.enterOuterAlt(localctx, 2)
                self.state = 104
                self.floating_point_type()
                self.state = 105
                self.identifier()
                self.state = 106
                self.match(RosMessageParser.ASSIGNMENT)
                self.state = 109
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 5, self._ctx)
                if la_ == 1:
                    self.state = 107
                    self.integral_value()
                    pass

                elif la_ == 2:
                    self.state = 108
                    self.floating_point_value()
                    pass

                pass
            elif token in [1]:
                self.enterOuterAlt(localctx, 3)
                self.state = 111
                self.boolean_type()
                self.state = 112
                self.identifier()
                self.state = 113
                self.match(RosMessageParser.ASSIGNMENT)
                self.state = 116
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [27, 28]:
                    self.state = 114
                    self.bool_value()
                    pass
                elif token in [21, 22, 25]:
                    self.state = 115
                    self.integral_value()
                    pass
                else:
                    raise NoViableAltException(self)

                pass
            elif token in [14]:
                self.enterOuterAlt(localctx, 4)
                self.state = 118
                self.string_type()
                self.state = 119
                self.identifier()
                self.state = 120
                self.match(RosMessageParser.STRING_ASSIGNMENT)
                self.state = 121
                self.string_value()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class CommentContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COMMENT(self):
            return self.getToken(RosMessageParser.COMMENT, 0)

        def HASH(self):
            return self.getToken(RosMessageParser.HASH, 0)

        def STRING_HASH(self):
            return self.getToken(RosMessageParser.STRING_HASH, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_comment

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterComment"):
                listener.enterComment(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitComment"):
                listener.exitComment(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitComment"):
                return visitor.visitComment(self)
            else:
                return visitor.visitChildren(self)

    def comment(self):

        localctx = RosMessageParser.CommentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_comment)
        self._la = 0  # Token type
        try:
            self.state = 128
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 8, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 125
                _la = self._input.LA(1)
                if not (_la == 23 or _la == 39):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 126
                self.match(RosMessageParser.COMMENT)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 127
                _la = self._input.LA(1)
                if not (_la == 23 or _la == 39):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class IdentifierContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(RosMessageParser.IDENTIFIER, 0)

        def STRING_IDENTIFIER(self):
            return self.getToken(RosMessageParser.STRING_IDENTIFIER, 0)

        def INT8(self):
            return self.getToken(RosMessageParser.INT8, 0)

        def UINT8(self):
            return self.getToken(RosMessageParser.UINT8, 0)

        def INT16(self):
            return self.getToken(RosMessageParser.INT16, 0)

        def UINT16(self):
            return self.getToken(RosMessageParser.UINT16, 0)

        def INT32(self):
            return self.getToken(RosMessageParser.INT32, 0)

        def UINT32(self):
            return self.getToken(RosMessageParser.UINT32, 0)

        def INT64(self):
            return self.getToken(RosMessageParser.INT64, 0)

        def UINT64(self):
            return self.getToken(RosMessageParser.UINT64, 0)

        def BYTE(self):
            return self.getToken(RosMessageParser.BYTE, 0)

        def CHAR(self):
            return self.getToken(RosMessageParser.CHAR, 0)

        def FLOAT32(self):
            return self.getToken(RosMessageParser.FLOAT32, 0)

        def FLOAT64(self):
            return self.getToken(RosMessageParser.FLOAT64, 0)

        def TIME(self):
            return self.getToken(RosMessageParser.TIME, 0)

        def DURATION(self):
            return self.getToken(RosMessageParser.DURATION, 0)

        def STRING(self):
            return self.getToken(RosMessageParser.STRING, 0)

        def BOOL(self):
            return self.getToken(RosMessageParser.BOOL, 0)

        def TRUE(self):
            return self.getToken(RosMessageParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(RosMessageParser.FALSE, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_identifier

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIdentifier"):
                listener.enterIdentifier(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIdentifier"):
                listener.exitIdentifier(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIdentifier"):
                return visitor.visitIdentifier(self)
            else:
                return visitor.visitChildren(self)

    def identifier(self):

        localctx = RosMessageParser.IdentifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_identifier)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 130
            _la = self._input.LA(1)
            if not ((((_la) & ~0x3F) == 0 and ((1 << _la) & 18119524350) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Type_Context(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def integral_type(self):
            return self.getTypedRuleContext(RosMessageParser.Integral_typeContext, 0)

        def floating_point_type(self):
            return self.getTypedRuleContext(
                RosMessageParser.Floating_point_typeContext, 0
            )

        def temporal_type(self):
            return self.getTypedRuleContext(RosMessageParser.Temporal_typeContext, 0)

        def boolean_type(self):
            return self.getTypedRuleContext(RosMessageParser.Boolean_typeContext, 0)

        def string_type(self):
            return self.getTypedRuleContext(RosMessageParser.String_typeContext, 0)

        def ros_type(self):
            return self.getTypedRuleContext(RosMessageParser.Ros_typeContext, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_type_

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterType_"):
                listener.enterType_(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitType_"):
                listener.exitType_(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitType_"):
                return visitor.visitType_(self)
            else:
                return visitor.visitChildren(self)

    def type_(self):

        localctx = RosMessageParser.Type_Context(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_type_)
        try:
            self.state = 138
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [2, 3, 4, 5, 6, 7, 8, 9, 10, 11]:
                self.enterOuterAlt(localctx, 1)
                self.state = 132
                self.integral_type()
                pass
            elif token in [12, 13]:
                self.enterOuterAlt(localctx, 2)
                self.state = 133
                self.floating_point_type()
                pass
            elif token in [15, 16]:
                self.enterOuterAlt(localctx, 3)
                self.state = 134
                self.temporal_type()
                pass
            elif token in [1]:
                self.enterOuterAlt(localctx, 4)
                self.state = 135
                self.boolean_type()
                pass
            elif token in [14]:
                self.enterOuterAlt(localctx, 5)
                self.state = 136
                self.string_type()
                pass
            elif token in [29]:
                self.enterOuterAlt(localctx, 6)
                self.state = 137
                self.ros_type()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ros_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self, i: int = None):
            if i is None:
                return self.getTokens(RosMessageParser.IDENTIFIER)
            else:
                return self.getToken(RosMessageParser.IDENTIFIER, i)

        def SLASH(self):
            return self.getToken(RosMessageParser.SLASH, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_ros_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRos_type"):
                listener.enterRos_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRos_type"):
                listener.exitRos_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRos_type"):
                return visitor.visitRos_type(self)
            else:
                return visitor.visitChildren(self)

    def ros_type(self):

        localctx = RosMessageParser.Ros_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_ros_type)
        try:
            self.state = 144
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 10, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 140
                self.match(RosMessageParser.IDENTIFIER)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 141
                self.match(RosMessageParser.IDENTIFIER)
                self.state = 142
                self.match(RosMessageParser.SLASH)
                self.state = 143
                self.match(RosMessageParser.IDENTIFIER)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Array_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def variable_array_type(self):
            return self.getTypedRuleContext(
                RosMessageParser.Variable_array_typeContext, 0
            )

        def fixed_array_type(self):
            return self.getTypedRuleContext(RosMessageParser.Fixed_array_typeContext, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_array_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArray_type"):
                listener.enterArray_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArray_type"):
                listener.exitArray_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitArray_type"):
                return visitor.visitArray_type(self)
            else:
                return visitor.visitChildren(self)

    def array_type(self):

        localctx = RosMessageParser.Array_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_array_type)
        try:
            self.state = 148
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 11, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 146
                self.variable_array_type()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 147
                self.fixed_array_type()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Variable_array_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def type_(self):
            return self.getTypedRuleContext(RosMessageParser.Type_Context, 0)

        def OPEN_BRACKET(self):
            return self.getToken(RosMessageParser.OPEN_BRACKET, 0)

        def STRING_OPEN_BRACKET(self):
            return self.getToken(RosMessageParser.STRING_OPEN_BRACKET, 0)

        def CLOSE_BRACKET(self):
            return self.getToken(RosMessageParser.CLOSE_BRACKET, 0)

        def STRING_CLOSE_BRACKET(self):
            return self.getToken(RosMessageParser.STRING_CLOSE_BRACKET, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_variable_array_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterVariable_array_type"):
                listener.enterVariable_array_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitVariable_array_type"):
                listener.exitVariable_array_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitVariable_array_type"):
                return visitor.visitVariable_array_type(self)
            else:
                return visitor.visitChildren(self)

    def variable_array_type(self):

        localctx = RosMessageParser.Variable_array_typeContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 28, self.RULE_variable_array_type)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 150
            self.type_()
            self.state = 151
            _la = self._input.LA(1)
            if not (_la == 18 or _la == 35):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 152
            _la = self._input.LA(1)
            if not (_la == 19 or _la == 36):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Fixed_array_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def type_(self):
            return self.getTypedRuleContext(RosMessageParser.Type_Context, 0)

        def OPEN_BRACKET(self):
            return self.getToken(RosMessageParser.OPEN_BRACKET, 0)

        def INTEGER_LITERAL(self):
            return self.getToken(RosMessageParser.INTEGER_LITERAL, 0)

        def CLOSE_BRACKET(self):
            return self.getToken(RosMessageParser.CLOSE_BRACKET, 0)

        def STRING_OPEN_BRACKET(self):
            return self.getToken(RosMessageParser.STRING_OPEN_BRACKET, 0)

        def STRING_INTEGER_LITERAL(self):
            return self.getToken(RosMessageParser.STRING_INTEGER_LITERAL, 0)

        def STRING_CLOSE_BRACKET(self):
            return self.getToken(RosMessageParser.STRING_CLOSE_BRACKET, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_fixed_array_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFixed_array_type"):
                listener.enterFixed_array_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFixed_array_type"):
                listener.exitFixed_array_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFixed_array_type"):
                return visitor.visitFixed_array_type(self)
            else:
                return visitor.visitChildren(self)

    def fixed_array_type(self):

        localctx = RosMessageParser.Fixed_array_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_fixed_array_type)
        try:
            self.state = 164
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 12, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 154
                self.type_()
                self.state = 155
                self.match(RosMessageParser.OPEN_BRACKET)
                self.state = 156
                self.match(RosMessageParser.INTEGER_LITERAL)
                self.state = 157
                self.match(RosMessageParser.CLOSE_BRACKET)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 159
                self.type_()
                self.state = 160
                self.match(RosMessageParser.STRING_OPEN_BRACKET)
                self.state = 161
                self.match(RosMessageParser.STRING_INTEGER_LITERAL)
                self.state = 162
                self.match(RosMessageParser.STRING_CLOSE_BRACKET)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Integral_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT8(self):
            return self.getToken(RosMessageParser.INT8, 0)

        def UINT8(self):
            return self.getToken(RosMessageParser.UINT8, 0)

        def INT16(self):
            return self.getToken(RosMessageParser.INT16, 0)

        def UINT16(self):
            return self.getToken(RosMessageParser.UINT16, 0)

        def INT32(self):
            return self.getToken(RosMessageParser.INT32, 0)

        def UINT32(self):
            return self.getToken(RosMessageParser.UINT32, 0)

        def INT64(self):
            return self.getToken(RosMessageParser.INT64, 0)

        def UINT64(self):
            return self.getToken(RosMessageParser.UINT64, 0)

        def BYTE(self):
            return self.getToken(RosMessageParser.BYTE, 0)

        def CHAR(self):
            return self.getToken(RosMessageParser.CHAR, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_integral_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIntegral_type"):
                listener.enterIntegral_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIntegral_type"):
                listener.exitIntegral_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIntegral_type"):
                return visitor.visitIntegral_type(self)
            else:
                return visitor.visitChildren(self)

    def integral_type(self):

        localctx = RosMessageParser.Integral_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_integral_type)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 166
            _la = self._input.LA(1)
            if not ((((_la) & ~0x3F) == 0 and ((1 << _la) & 4092) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Floating_point_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FLOAT32(self):
            return self.getToken(RosMessageParser.FLOAT32, 0)

        def FLOAT64(self):
            return self.getToken(RosMessageParser.FLOAT64, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_floating_point_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFloating_point_type"):
                listener.enterFloating_point_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFloating_point_type"):
                listener.exitFloating_point_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFloating_point_type"):
                return visitor.visitFloating_point_type(self)
            else:
                return visitor.visitChildren(self)

    def floating_point_type(self):

        localctx = RosMessageParser.Floating_point_typeContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 34, self.RULE_floating_point_type)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 168
            _la = self._input.LA(1)
            if not (_la == 12 or _la == 13):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Temporal_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TIME(self):
            return self.getToken(RosMessageParser.TIME, 0)

        def DURATION(self):
            return self.getToken(RosMessageParser.DURATION, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_temporal_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTemporal_type"):
                listener.enterTemporal_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTemporal_type"):
                listener.exitTemporal_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTemporal_type"):
                return visitor.visitTemporal_type(self)
            else:
                return visitor.visitChildren(self)

    def temporal_type(self):

        localctx = RosMessageParser.Temporal_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_temporal_type)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 170
            _la = self._input.LA(1)
            if not (_la == 15 or _la == 16):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class String_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(RosMessageParser.STRING, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_string_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterString_type"):
                listener.enterString_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitString_type"):
                listener.exitString_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitString_type"):
                return visitor.visitString_type(self)
            else:
                return visitor.visitChildren(self)

    def string_type(self):

        localctx = RosMessageParser.String_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_string_type)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 172
            self.match(RosMessageParser.STRING)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Boolean_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BOOL(self):
            return self.getToken(RosMessageParser.BOOL, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_boolean_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBoolean_type"):
                listener.enterBoolean_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBoolean_type"):
                listener.exitBoolean_type(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBoolean_type"):
                return visitor.visitBoolean_type(self)
            else:
                return visitor.visitChildren(self)

    def boolean_type(self):

        localctx = RosMessageParser.Boolean_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_boolean_type)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 174
            self.match(RosMessageParser.BOOL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SignContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def PLUS(self):
            return self.getToken(RosMessageParser.PLUS, 0)

        def MINUS(self):
            return self.getToken(RosMessageParser.MINUS, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_sign

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSign"):
                listener.enterSign(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSign"):
                listener.exitSign(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitSign"):
                return visitor.visitSign(self)
            else:
                return visitor.visitChildren(self)

    def sign(self):

        localctx = RosMessageParser.SignContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_sign)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 176
            _la = self._input.LA(1)
            if not (_la == 21 or _la == 22):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Integral_valueContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INTEGER_LITERAL(self):
            return self.getToken(RosMessageParser.INTEGER_LITERAL, 0)

        def sign(self):
            return self.getTypedRuleContext(RosMessageParser.SignContext, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_integral_value

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIntegral_value"):
                listener.enterIntegral_value(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIntegral_value"):
                listener.exitIntegral_value(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIntegral_value"):
                return visitor.visitIntegral_value(self)
            else:
                return visitor.visitChildren(self)

    def integral_value(self):

        localctx = RosMessageParser.Integral_valueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 44, self.RULE_integral_value)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 179
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 21 or _la == 22:
                self.state = 178
                self.sign()

            self.state = 181
            self.match(RosMessageParser.INTEGER_LITERAL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Floating_point_valueContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def REAL_LITERAL(self):
            return self.getToken(RosMessageParser.REAL_LITERAL, 0)

        def sign(self):
            return self.getTypedRuleContext(RosMessageParser.SignContext, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_floating_point_value

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFloating_point_value"):
                listener.enterFloating_point_value(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFloating_point_value"):
                listener.exitFloating_point_value(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFloating_point_value"):
                return visitor.visitFloating_point_value(self)
            else:
                return visitor.visitChildren(self)

    def floating_point_value(self):

        localctx = RosMessageParser.Floating_point_valueContext(
            self, self._ctx, self.state
        )
        self.enterRule(localctx, 46, self.RULE_floating_point_value)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 184
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == 21 or _la == 22:
                self.state = 183
                self.sign()

            self.state = 186
            self.match(RosMessageParser.REAL_LITERAL)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Bool_valueContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TRUE(self):
            return self.getToken(RosMessageParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(RosMessageParser.FALSE, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_bool_value

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBool_value"):
                listener.enterBool_value(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBool_value"):
                listener.exitBool_value(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBool_value"):
                return visitor.visitBool_value(self)
            else:
                return visitor.visitChildren(self)

    def bool_value(self):

        localctx = RosMessageParser.Bool_valueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 48, self.RULE_bool_value)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 188
            _la = self._input.LA(1)
            if not (_la == 27 or _la == 28):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class String_valueContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING_VALUE(self):
            return self.getToken(RosMessageParser.STRING_VALUE, 0)

        def getRuleIndex(self):
            return RosMessageParser.RULE_string_value

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterString_value"):
                listener.enterString_value(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitString_value"):
                listener.exitString_value(self)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitString_value"):
                return visitor.visitString_value(self)
            else:
                return visitor.visitChildren(self)

    def string_value(self):

        localctx = RosMessageParser.String_valueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 50, self.RULE_string_value)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 190
            self.match(RosMessageParser.STRING_VALUE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



from lqs.interface.core import ListInterface
from lqs.client.common import RESTInterface
import lqs.interface.core.models as models


class List(ListInterface, RESTInterface):
    service: str = "lqs"

    def __init__(self, app):
        super().__init__(app=app)

    def _api_key(self, **params):
        resource_path = "apiKeys" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.APIKeyListResponse
        )
        return result

    def _callback(self, **params):
        resource_path = "callbacks" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.CallbackListResponse
        )
        return result

    def _digestion(self, **params):
        resource_path = "digestions" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.DigestionListResponse
        )
        return result

    def _digestion_part(self, **params):
        digestion_id = params.pop("digestion_id", None)
        if digestion_id is None:
            url = "digestionParts"
        else:
            url = f"digestions/{digestion_id}/parts"
        resource_path = url + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.DigestionPartListResponse
        )
        return result

    def _digestion_topic(self, **params):
        digestion_id = params.pop("digestion_id", None)
        if digestion_id is None:
            url = "digestionTopics"
        else:
            url = f"digestions/{digestion_id}/topics"
        resource_path = url + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.DigestionTopicListResponse
        )
        return result

    def _group(self, **params):
        resource_path = "groups" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.GroupListResponse
        )
        return result

    def _hook(self, **params):
        workflow_id = params.pop("workflow_id", None)
        if workflow_id is None:
            url = "hooks"
        else:
            url = f"workflows/{workflow_id}/hooks"
        resource_path = url + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.HookListResponse
        )
        return result

    def _ingestion(self, **params):
        resource_path = "ingestions" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.IngestionListResponse
        )
        return result

    def _ingestion_part(self, **params):
        ingestion_id = params.pop("ingestion_id", None)
        if ingestion_id is None:
            url = "ingestionParts"
        else:
            url = f"ingestions/{ingestion_id}/parts"
        resource_path = url + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.IngestionPartListResponse
        )
        return result

    def _label(self, **params):
        resource_path = "labels" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.LabelListResponse
        )
        return result

    def _log(self, **params):
        resource_path = "logs" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.LogListResponse
        )
        return result

    def _log_object(self, **params):
        log_id = params.pop("log_id")
        resource_path = f"logs/{log_id}/objects" + self._get_url_param_string(
            params, []
        )
        result = self._get_resource(
            resource_path, response_model=models.ObjectListResponse
        )
        return result

    def _log_object_part(self, **params):
        log_id = params.pop("log_id")
        object_key = params.pop("object_key")
        resource_path = (
            f"logs/{log_id}/objects/{object_key}/parts"
            + self._get_url_param_string(params, [])
        )
        result = self._get_resource(
            resource_path, response_model=models.ObjectPartListResponse
        )
        return result

    def _object(self, **params):
        object_store_id = params.pop("object_store_id")
        resource_path = (
            f"objectStores/{object_store_id}/objects"
            + self._get_url_param_string(params, [])
        )
        result = self._get_resource(
            resource_path, response_model=models.ObjectListResponse
        )
        return result

    def _object_part(self, **params):
        raise NotImplementedError

    def _object_store(self, **params):
        resource_path = "objectStores" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.ObjectStoreListResponse
        )
        return result

    def _query(self, **params):
        log_id = params.pop("log_id", None)
        if log_id is None:
            url = "queries"
        else:
            url = f"logs/{log_id}/queries"
        resource_path = url + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.QueryListResponse
        )
        return result

    def _record(self, **params):
        topic_id = params.pop("topic_id")
        resource_path = f"topics/{topic_id}/records" + self._get_url_param_string(
            params, []
        )
        result = self._get_resource(
            resource_path, response_model=models.RecordListResponse
        )
        return result

    def _role(self, **params):
        resource_path = "roles" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.RoleListResponse
        )
        return result

    def _tag(self, **params):
        log_id = params.pop("log_id", None)
        if log_id is None:
            url = "tags"
        else:
            url = f"logs/{log_id}/tags"
        resource_path = url + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.TagListResponse
        )
        return result

    def _topic(self, **params):
        resource_path = "topics" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.TopicListResponse
        )
        return result

    def _user(self, **params):
        resource_path = "users" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.UserListResponse
        )
        return result

    def _workflow(self, **params):
        resource_path = "workflows" + self._get_url_param_string(params, [])
        result = self._get_resource(
            resource_path, response_model=models.WorkflowListResponse
        )
        return result



from typing import Optional

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.base.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)
from lqs.interface.core import models


class ObjectStore(CommonModel):
    name: Optional[str] = Field(
        None, description="The unique name of the object store for reference."
    )
    bucket_name: str = Field(
        ..., description="The name of the bucket for the object store."
    )
    access_key_id: Optional[str] = Field(
        None, description="The access key ID used to access the object store's bucket."
    )
    region_name: Optional[str] = Field(
        None,
        description="The name of the region where the object store's bucket is located.",
    )
    endpoint_url: Optional[str] = Field(
        None,
        description="The endpoint URL of the object store's bucket. If not provided, the AWS default endpoint will be used.",
    )
    note: Optional[str] = Field(
        None, description="A general note about the object store for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the object store."
    )
    disabled: bool = Field(
        False, description="Whether the object store is disabled and cannot be used."
    )
    default: bool = Field(
        False,
        description="Whether the object store is the default object store. Used for reference.",
    )
    read_only: bool = Field(
        False,
        description="Whether the object store can only be used for read operations.",
    )
    managed: bool = Field(
        False, description="Whether the object store is managed by the system."
    )
    key_prefix: Optional[str] = Field(
        None,
        description="The prefix for enforcing/limiting the object store's objects' keys.",
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            [
                "name",
                "access_key_id",
                "region_name",
                "endpoint_url",
                "note",
                "context",
                "disabled",
                "default",
                "read_only",
                "managed",
                "key_prefix",
            ]
        )
    )

    def list_objects(self, threaded=False, **kwargs) -> list["models.Object"]:
        return self._list_all_subresources(
            list_method=self._app.list.object,
            threaded=threaded,
            object_store_id=self.id,
            **kwargs,
        )


class ObjectStoreDataResponse(DataResponseModel[ObjectStore]):
    pass


class ObjectStoreListResponse(PaginationModel[ObjectStore]):
    pass


class ObjectStoreCreateRequest(BaseModel):
    name: str
    bucket_name: str
    access_key_id: Optional[str] = None
    secret_access_key: Optional[str] = None
    region_name: Optional[str] = None
    endpoint_url: Optional[str] = None
    note: Optional[str] = None
    context: Optional[dict] = None
    disabled: bool = False
    default: bool = False
    read_only: bool = False
    managed: bool = False
    key_prefix: Optional[str] = None


class ObjectStoreUpdateRequest(BaseModel):
    name: str = optional_field
    bucket_name: str = optional_field
    access_key_id: Optional[str] = optional_field
    secret_access_key: Optional[str] = optional_field
    region_name: Optional[str] = optional_field
    endpoint_url: Optional[str] = optional_field
    note: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    disabled: bool = optional_field
    default: bool = optional_field
    read_only: bool = optional_field
    managed: bool = optional_field
    key_prefix: Optional[str] = optional_field



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lqs.client import RESTClient
from lqs.client.dsm.list import List
from lqs.client.dsm.fetch import Fetch
from lqs.client.dsm.create import Create
from lqs.client.dsm.update import Update
from lqs.client.dsm.delete import Delete


class DataStoreManager:
    def __init__(self, app: "RESTClient"):
        self.app = app

        self.list = List(app=app)
        self.fetch = Fetch(app=app)
        self.create = Create(app=app)
        self.update = Update(app=app)
        self.delete = Delete(app=app)



import io
from typing import Union

import orjson as json

from lqs.common.utils import get_logger
from lqs.transcode.cdr import CDRWrapper
from lqs.transcode.ark.ark_json_parser import RbufMessageDeserializer
from lqs.transcode.config import TranscodeConfig
from lqs.transcode.cutter import CutterDecoder
from lqs.transcode.ros1.ros_message_deserializer import RosMessageDeserializer


class Transcode:
    def __init__(self, config: Union[TranscodeConfig, dict] = TranscodeConfig()):
        self.config = (
            config if isinstance(config, TranscodeConfig) else TranscodeConfig(**config)
        )
        self.ros_deserializer = RosMessageDeserializer(
            trim_size=self.config.trim_cutoff
        )
        self.rbuf_deserializer = RbufMessageDeserializer(
            trim_size=self.config.trim_cutoff
        )
        self.cdr_wrapper = CDRWrapper()

        self.cutter_decoder = CutterDecoder()
        self.logger = get_logger(
            self.__class__.__name__,
            level=self.config.log_level,
            json_logging=self.config.log_as_json,
        )

    @classmethod
    def supported_encoding(cls):
        return ["ros1", "rbuf", "json", "cdr"]

    def deserialize(
        self,
        type_encoding,
        type_name,
        type_data,
        message_bytes,
        replacements=((b"\\u0000", b"NULL"),),
    ):
        if type_encoding == "ros1":
            message_bytes = io.BytesIO(message_bytes)
            res = self.ros_deserializer.deserialize(
                message_type=type_name,
                message_bytes=message_bytes,
                message_type_data=type_data,
            )
            res_dump = json.dumps(res, default=lambda x: x.to_dict())
            for replacement in replacements:
                res_dump = res_dump.replace(*replacement)
            return json.loads(res_dump)

        if type_encoding == "rbuf":
            message_bytes = io.BytesIO(message_bytes)
            res = self.rbuf_deserializer.deserialize(
                message_bytes=message_bytes,
                message_type=type_name,
                message_type_data=json.loads(type_data),
            )
            res_dump = json.dumps(res, default=lambda x: x.to_dict())
            for replacement in replacements:
                res_dump = res_dump.replace(*replacement)
            return json.loads(res_dump)

        if type_encoding == "json":
            return json.loads(message_bytes)

        if type_encoding == "cdr":
            return self.cdr_wrapper.deserialize(
                type_name=type_name,
                type_data=type_data,
                message_bytes=message_bytes,
            )

        if not type_encoding and type_name and type_name.startswith("lqs_cutter"):
            return self.cutter_decoder.deserialize(
                type_name=type_name, payload=message_bytes
            )

        if type_encoding is None:
            return message_bytes

        raise NotImplementedError(f"Unknown type encoding: {type_encoding}")

    def serialize(self, type_encoding, type_name, type_data, message_dict):
        pass

    def get_schema(self, type_encoding, type_name, type_data) -> dict:
        if type_encoding == "ros1":
            res = self.ros_deserializer.get_json_schema(
                message_type=type_name, message_type_data=type_data
            )
        elif type_encoding == "rbuf":
            res = self.rbuf_deserializer.get_json_schema(
                message_type=type_name, message_type_data=type_data
            )
        else:
            raise NotImplementedError(f"Unknown type encoding: {type_encoding}")

        return res



from typing import List, Optional, Dict, Any
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)


class Query(CommonModel):
    log_id: UUID = Field(
        ..., description="The ID of the log to which the query is scoped."
    )
    name: Optional[str] = Field(
        None, description="The name of the query for reference (not unique)."
    )
    note: Optional[str] = Field(
        None, description="A general note about the query for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the query."
    )

    statement: Optional[str] = Field(
        None, description="The SQL statement of the query."
    )
    parameters: Optional[Dict[str, Any]] = Field(
        None, description="The parameters of the query."
    )
    columns: Optional[List[str]] = Field(
        None, description="The columns of the query result."
    )
    rows: Optional[List[List[Any]]] = Field(
        None, description="The rows of the query result."
    )
    error: Optional[dict] = Field(None, description="The error of the query if any.")

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            [
                "name",
                "note",
                "context",
                "statement",
                "parameters",
                "columns",
                "rows",
                "error",
            ]
        )
    )


class QueryDataResponse(DataResponseModel[Query]):
    pass


class QueryListResponse(PaginationModel[Query]):
    pass


class QueryForLogCreateRequest(BaseModel):
    name: Optional[str] = None
    note: Optional[str] = None
    context: Optional[dict] = None
    statement: Optional[str] = None
    parameters: Optional[Dict[str, Any]] = None


class QueryCreateRequest(QueryForLogCreateRequest):
    log_id: UUID


class QueryUpdateRequest(BaseModel):
    name: Optional[str] = optional_field
    note: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    statement: Optional[str] = optional_field
    parameters: Optional[Dict[str, Any]] = optional_field



from abc import abstractmethod
from uuid import UUID

import lqs.interface.dsm.models as models
from lqs.interface.base.update import UpdateInterface as BaseUpdateInterface


class UpdateInterface(BaseUpdateInterface):
    @abstractmethod
    def _announcement(self, **kwargs) -> models.AnnouncementDataResponse:
        pass

    def announcement(
        self,
        announcement_id: UUID,
        data: models.AnnouncementUpdateRequest,
    ):
        return self._announcement(
            announcement_id=announcement_id,
            data=self._process_data(data),
        )

    def _announcement_by_model(
        self, announcement_id: UUID, data: models.AnnouncementUpdateRequest
    ):
        return self.announcement(announcement_id=announcement_id, data=data)

    @abstractmethod
    def _comment(self, **kwargs) -> models.CommentDataResponse:
        pass

    def comment(
        self,
        comment_id: UUID,
        data: models.CommentUpdateRequest,
    ):
        return self._comment(
            comment_id=comment_id,
            data=self._process_data(data),
        )

    def _comment_by_model(self, comment_id: UUID, data: models.CommentUpdateRequest):
        return self.comment(comment_id=comment_id, data=data)

    @abstractmethod
    def _configuration(self, **kwargs) -> models.ConfigurationDataResponse:
        pass

    def configuration(
        self,
        configuration_id: UUID,
        data: models.ConfigurationUpdateRequest,
    ):
        return self._configuration(
            configuration_id=configuration_id,
            data=self._process_data(data),
        )

    def _configuration_by_model(
        self, configuration_id: UUID, data: models.ConfigurationUpdateRequest
    ):
        return self.configuration(configuration_id=configuration_id, data=data)

    @abstractmethod
    def _datastore(self, **kwargs) -> models.DataStoreDataResponse:
        pass

    def datastore(self, datastore_id: UUID, data: models.DataStoreUpdateRequest):
        return self._datastore(
            datastore_id=datastore_id,
            data=self._process_data(data),
        )

    @abstractmethod
    def _datastore_association(
        self, **kwargs
    ) -> models.DataStoreAssociationDataResponse:
        pass

    def datastore_association(
        self,
        datastore_association_id: UUID,
        data: models.DataStoreAssociationUpdateRequest,
    ):
        return self._datastore_association(
            datastore_association_id=datastore_association_id,
            data=self._process_data(data),
        )

    @abstractmethod
    def _event(self, **kwargs) -> models.EventDataResponse:
        pass

    def event(self, event_id: UUID, data: models.EventUpdateRequest):
        return self._event(
            event_id=event_id,
            data=self._process_data(data),
        )

    @abstractmethod
    def _job(self, **kwargs) -> models.JobDataResponse:
        pass

    def job(self, job_id: UUID, data: models.JobUpdateRequest):
        return self._job(
            job_id=job_id,
            data=self._process_data(data),
        )

    @abstractmethod
    def _user(self, **kwargs) -> models.UserDataResponse:
        pass

    def user(
        self, user_id: UUID, data: models.UserUpdateRequest
    ) -> models.UserDataResponse:
        return self._user(user_id=user_id, data=self._process_data(data))



# read functions for primitive types
import struct
from typing import Dict

from lqs.transcode.errors import DeserializationError

numerical_types = {
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "float",
    "double",
}
fixed_length_type = {
    *numerical_types,
    "bool",
    "guid",
    "duration",
    "steady_time_point",
    "system_time_point",
    "enum",
}

# The magic bits to indicate the byte is a header.
BITSTREAM_HEADER_MAGIC = 0xD0

# THe magic bits to indicate the byte is a group.
BITSTREAM_GROUP_MAGIC = 0xE0

# Indicates that this header/group has at least one following section
# that needs to be read.
BITSTREAM_HAS_FOLLOWING_SECTION = 0x04


def read_int8(stream, TRIM_SIZE):
    return struct.unpack("<b", stream.read(1))[0]


def read_uint8(stream, TRIM_SIZE):
    return struct.unpack("<B", stream.read(1))[0]


def read_int16(stream, TRIM_SIZE):
    return struct.unpack("<h", stream.read(2))[0]


def read_uint16(stream, TRIM_SIZE):
    return struct.unpack("<H", stream.read(2))[0]


def read_int32(stream, TRIM_SIZE):
    return struct.unpack("<i", stream.read(4))[0]


def read_uint32(stream, TRIM_SIZE):
    return struct.unpack("<I", stream.read(4))[0]


def read_int64(stream, TRIM_SIZE):
    return struct.unpack("<q", stream.read(8))[0]


def read_uint64(stream, TRIM_SIZE):
    return struct.unpack("<Q", stream.read(8))[0]


def read_float(stream, TRIM_SIZE):
    return struct.unpack("<f", stream.read(4))[0]


def read_double(stream, TRIM_SIZE):
    return struct.unpack("<d", stream.read(8))[0]


def read_string(stream, TRIM_SIZE):
    length = read_uint32(stream, TRIM_SIZE=TRIM_SIZE)
    if TRIM_SIZE is not None and length > TRIM_SIZE:
        stream.read(length)
        return None

    return stream.read(length).decode("utf-8")


def read_time(stream, TRIM_SIZE):
    return read_uint64(stream, TRIM_SIZE=TRIM_SIZE)


def read_steady_time_point(stream, TRIM_SIZE):
    return read_time(stream, TRIM_SIZE=TRIM_SIZE)


def read_system_time_point(stream, TRIM_SIZE):
    return read_time(stream, TRIM_SIZE=TRIM_SIZE)


def read_duration(stream, TRIM_SIZE):
    return read_uint64(stream, TRIM_SIZE=TRIM_SIZE)


def read_bool(stream, TRIM_SIZE):
    return read_uint8(stream, TRIM_SIZE=TRIM_SIZE) != 0


def read_guid(stream, TRIM_SIZE):
    # TODO revisit this later
    """
    Reads a GUID from the bitstream, which is essentially a 16-byte array that we
    return into a string for Python consumption.
    """

    # Read the raw 16 bytes in.
    raw_bytes = stream.read(16)

    # Convert the array into a hex guid, the hard way. Note that in Ark
    # the first 8 bytes are behind the second 8 bytes, and 'backwards', as
    # it tries to match little endian notation for two 8 byte integers for
    # some reason. We should move this into a class.
    output = ""

    index = 7
    while index >= 0:
        if index == 1 or index == 3:
            output += "-"

        output += "{:02x}".format(raw_bytes[index])
        index -= 1

    index = 15
    while index >= 8:
        if index == 15 or index == 13:
            output += "-"

        output += "{:02x}".format(raw_bytes[index])
        index -= 1

    return output


def read_byte_buffer(stream, TRIM_SIZE):
    length = read_uint32(stream, TRIM_SIZE=TRIM_SIZE)

    result = stream.read(length)
    if TRIM_SIZE is not None and length > TRIM_SIZE:
        return None
    # convert byte buffer to list to make it json compatible (for db insert)
    return list(result)


def read_array(type, stream, read_element, TRIM_SIZE, length=None):
    if length is None:
        length = read_uint32(stream, TRIM_SIZE=TRIM_SIZE)

    if type in fixed_length_type:
        # TODO can optimize this
        pass
    result = [read_element(stream, TRIM_SIZE=TRIM_SIZE) for _ in range(length)]
    if TRIM_SIZE is not None and length > TRIM_SIZE:
        return None
    return result


def read_dictionary(stream, read_key, read_value, TRIM_SIZE, length=None):
    if length is None:
        length = read_uint32(stream, TRIM_SIZE=TRIM_SIZE)
    return {
        str(read_key(stream, TRIM_SIZE=TRIM_SIZE)): read_value(stream, TRIM_SIZE=TRIM_SIZE)
        for _ in range(length)
    }


def consume(stream, length):
    # TODO maybe just seek instead depending on the stream
    stream.read(length)


def read_variant(stream, types: Dict, TRIM_SIZE):
    index = read_uint8(stream, TRIM_SIZE=TRIM_SIZE)
    length = read_uint32(stream, TRIM_SIZE=TRIM_SIZE)

    if index not in types:
        # consume the bytes
        consume(stream, length)
        return None

    return types[index]().deserialize(stream, TRIM_SIZE=TRIM_SIZE)


def read_bitstream_header(stream, TRIM_SIZE):
    """
    Reads the bitstream header from the stream. These headers
    denote the start of a new serialized structure, and help
    indicate version information or if different sections are
    present.
    """

    header = read_uint8(stream, TRIM_SIZE=TRIM_SIZE)

    # Ensure the header magic bits are set.
    if (header & 0xF0) != BITSTREAM_HEADER_MAGIC:
        raise DeserializationError(
            "Tried to read a bitstream header, but the header had the incorrect magic numbers. "
            "This indicates the stream is corrupt, or is not a rbuf stream."
        )

    # We can only decode version one with this code.
    if (header & 0x03) != 0x01:
        raise DeserializationError(
            "Tried to read a bitstream header, but it had an unexpected version number. This "
            "software only supports version 1."
        )

    # Store off if we have more sections...
    has_more_sections = (header & BITSTREAM_HAS_FOLLOWING_SECTION) > 0

    # If the last bit is set, then we have some kind of unexpected
    # bitstream error.
    if (header & 0x08) != 0x00:
        raise DeserializationError(
            "Tried to read a bitstream header, but this header indicates the stream has groups, "
            "which this software does not support."
        )
    return has_more_sections


def read_optional_group_header(stream, TRIM_SIZE):
    """
    Reads a group header from the bitstream, updating if we have additional sections or not,
    and returning a tuple containing the parsed identifier and group size. Throws if the next
    byte does not appear to be a group header.
    """

    header = struct.Struct("<BBL").unpack(stream.read(6))

    # Throw if this isn't a group
    if (header[0] & 0xF0) != BITSTREAM_GROUP_MAGIC:
        raise DeserializationError(
            "Tried to read a bitstream group header, but the header had an incorrect magic "
            "number. This indicates the stream is corrupt, or is not an rbuf stream."
        )

    # Record if we have additional sections
    has_more_sections = (header[0] & BITSTREAM_HAS_FOLLOWING_SECTION) > 0

    # Throw if we have any other fields set, something is wrong and we're an incorrect
    # version or were wrong that it was a group.
    if (header[0] & 0x0B) != 0x00:
        raise DeserializationError(
            "Tried to read a bitstream group header, but the header had additional flags specified that were unknown."
        )

    # Return the group identifier and size as a tuple
    return has_more_sections, header[1:]


def split_type_mapping_info_fn(stream, TRIM_SIZE):
    read_bitstream_header(stream, None)
    return {
        "object_identifier": read_guid(stream, None),
        "short_identifier": read_uint8(stream, None),
    }


def split_file_header_fn(stream):
    read_bitstream_header(stream, None)
    return {
        "format_version": read_uint16(stream, None),
        "split_identifier": read_guid(stream, None),
        "index_offset": read_uint32(stream, None),
        "indexed_item_count": read_uint32(stream, None),
        "maximum_item_count": read_uint32(stream, None),
        "type_mappings": read_array(
            type=None,
            stream=stream,
            read_element=lambda stream, TRIM_SIZE: split_type_mapping_info_fn(
                stream, None
            ),
            TRIM_SIZE=None,
        ),
    }



from abc import abstractmethod
from uuid import UUID

from lqs.interface.base.delete import DeleteInterface as BaseDeleteInterface


class DeleteInterface(BaseDeleteInterface):
    @abstractmethod
    def _announcement(self, **kwargs):
        pass

    def announcement(self, announcement_id: UUID):
        return self._announcement(
            announcement_id=announcement_id,
        )

    @abstractmethod
    def _comment(self, **kwargs):
        pass

    def comment(self, comment_id: UUID):
        return self._comment(
            comment_id=comment_id,
        )

    @abstractmethod
    def _configuration(self, **kwargs):
        pass

    def configuration(self, configuration_id: UUID):
        return self._configuration(
            configuration_id=configuration_id,
        )

    @abstractmethod
    def _datastore(self, **kwargs):
        pass

    def datastore(self, datastore_id: UUID):
        return self._datastore(
            datastore_id=datastore_id,
        )

    @abstractmethod
    def _datastore_association(self, **kwargs):
        pass

    def datastore_association(self, datastore_association_id: UUID):
        return self._datastore_association(
            datastore_association_id=datastore_association_id,
        )

    @abstractmethod
    def _event(self, **kwargs):
        pass

    def event(self, event_id: UUID):
        return self._event(
            event_id=event_id,
        )

    @abstractmethod
    def _job(self, **kwargs):
        pass

    def job(self, job_id: UUID):
        return self._job(
            job_id=job_id,
        )

    @abstractmethod
    def _user(self, **kwargs):
        pass

    def user(self, user_id: UUID):
        return self._user(
            user_id=user_id,
        )



from typing import Optional
from uuid import UUID

from pydantic import BaseModel

from lqs.interface.base.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)
from lqs.interface.base import models


class User(CommonModel):
    _repr_fields = ("id", "username")

    username: str
    role_id: Optional[UUID]
    admin: bool
    disabled: bool
    managed: bool
    external_id: Optional[str]

    def fetch_role(self) -> Optional["models.Role"]:
        if self.role_id is None:
            return None
        return self._app.fetch.role(self.role_id).data

    def list_api_keys(self) -> "models.APIKeyListResponse":
        return self._app.fetch.api_keys(user_id=self.id)


class UserDataResponse(BaseModel):
    data: User


class MeDataResponse(DataResponseModel[User]):
    data: Optional[User]


class UserListResponse(PaginationModel[User]):
    pass


class UserCreateRequest(BaseModel):
    username: str
    role_id: Optional[UUID] = None
    admin: bool = False
    disabled: bool = False
    managed: bool = False
    external_id: Optional[str] = None
    password: Optional[str] = None  # note: this is virtual


class UserUpdateRequest(BaseModel):
    username: str = optional_field
    role_id: Optional[UUID] = optional_field
    admin: bool = optional_field
    disabled: bool = optional_field
    external_id: Optional[str] = optional_field
    password: Optional[str] = optional_field  # note: this is virtual



from typing import List, Optional, Tuple, Union
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    DataResponseModel,
    PaginationModel,
    ProcessModel,
    ProcessCreateRequest,
    ProcessUpdateRequest,
    optional_field,
    Int64,
    process_required_fields,
)
from lqs.interface.core import models


class IngestionPartIndexEntry(BaseModel):
    topic_id: str
    data_offset: int
    data_length: int
    chunk_compression: Optional[str]
    chunk_offset: Optional[int]
    chunk_length: Optional[int]
    timestamp: Int64
    # context: Optional[dict] = None


# TODO: in Python 3.11, we can use Tuple[*get_type_hints(IngestionPartIndex).values()]
IngestionPartIndexTuple = Tuple[
    str, int, int, Optional[str], Optional[int], Optional[int], Int64
]

IngestionPartIndexTupleV2 = Tuple[
    str, int, int, Optional[str], Optional[int], Optional[int], Int64, Optional[dict]
]


class IngestionPartIndex(BaseModel):
    ingestion_part_id: UUID
    index: Optional[
        Union[List[IngestionPartIndexTuple], List[IngestionPartIndexTupleV2]]
    ]


class IngestionPartIndexCreateRequest(BaseModel):
    ingestion_part_id: UUID
    index: Optional[
        Union[List[IngestionPartIndexTuple], List[IngestionPartIndexTupleV2]]
    ] = None


class IngestionPartIndexUpdateRequest(BaseModel):
    index: Optional[
        Union[List[IngestionPartIndexTuple], List[IngestionPartIndexTupleV2]]
    ] = optional_field


class IngestionPartIndexDataResponse(DataResponseModel[IngestionPartIndex]):
    pass


class IngestionPartIndexListResponse(PaginationModel[IngestionPartIndex]):
    pass


# Ingestion Part


class IngestionPart(ProcessModel["IngestionPart"]):
    sequence: int = Field(
        ...,
        description="The index of the ingestion part within the ingestion for a given source.",
    )
    ingestion_id: UUID = Field(
        ..., description="The ID of the ingestion to which the ingestion part belongs."
    )
    source: Optional[str] = Field(
        None,
        description="The path of the source object relative to the ingestion's object key.",
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the ingestion part."
    )
    index: Optional[
        Union[List[IngestionPartIndexTuple], List[IngestionPartIndexTupleV2]]
    ] = Field(None, description="The record metadata index of the ingestion part.")

    log_id: Optional[UUID] = Field(
        None, description="The ID of the log to which the ingestion part belongs."
    )
    group_id: Optional[UUID] = Field(
        None, description="The ID of the group to which the ingestion part belongs."
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            ["source", "context", "index", "log_id", "group_id"]
            + process_required_fields
        )
    )

    def fetch_ingestion(self) -> "models.Ingestion":
        return self._app.fetch.ingestion(self.ingestion_id).data

    def fetch_log(self) -> "models.Log":
        return self._app.fetch.log(self.log_id).data

    def fetch_group(self) -> "models.Group":
        return self._app.fetch.group(self.group_id).data


class IngestionPartListResponse(PaginationModel[IngestionPart]):
    pass


class IngestionPartDataResponse(DataResponseModel[IngestionPart]):
    pass


class IngestionPartForIngestionCreateRequest(ProcessCreateRequest):
    sequence: int
    source: Optional[str] = None
    context: Optional[dict] = None
    index: Optional[
        Union[List[IngestionPartIndexTuple], List[IngestionPartIndexTupleV2]]
    ] = None


class IngestionPartCreateRequest(IngestionPartForIngestionCreateRequest):
    ingestion_id: UUID


class IngestionPartUpdateRequest(ProcessUpdateRequest):
    sequence: int = optional_field
    source: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    index: Optional[
        Union[List[IngestionPartIndexTuple], List[IngestionPartIndexTupleV2]]
    ] = optional_field



from typing import Optional
from uuid import UUID

from pydantic import BaseModel

from lqs.interface.dsm.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)


class DataStoreAssociation(CommonModel):
    user_id: UUID
    datastore_id: UUID
    manager: bool
    disabled: bool

    datastore_user_id: UUID
    datastore_username: Optional[str]
    datastore_role_id: Optional[UUID]
    datastore_admin: bool
    datastore_disabled: bool


class DataStoreAssociationDataResponse(DataResponseModel[DataStoreAssociation]):
    pass


class DataStoreAssociationListResponse(PaginationModel[DataStoreAssociation]):
    pass


class DataStoreAssociationCreateRequest(BaseModel):
    user_id: UUID
    datastore_id: UUID
    manager: bool = False
    disabled: bool = False

    datastore_user_id: Optional[UUID] = None
    datastore_username: Optional[str] = None
    datastore_role_id: Optional[UUID] = None
    datastore_admin: bool = False
    datastore_disabled: bool = False


class DataStoreAssociationUpdateRequest(BaseModel):
    manager: bool = optional_field
    disabled: bool = optional_field

    datastore_username: Optional[str] = optional_field
    datastore_role_id: Optional[UUID] = optional_field
    datastore_admin: bool = optional_field
    datastore_disabled: bool = optional_field



from typing import Optional
from uuid import UUID

from pydantic import BaseModel

from lqs.interface.dsm.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    ProcessState,
    ProcessType,
    JobType,
    optional_field,
)


class Job(CommonModel):
    event_id: Optional[UUID]
    type: JobType
    process_type: ProcessType
    resource_id: UUID
    datastore_id: Optional[UUID]
    datastore_endpoint: Optional[str]
    state: ProcessState
    error: Optional[dict]


class JobDataResponse(DataResponseModel[Job]):
    pass


class JobListResponse(PaginationModel[Job]):
    pass


class JobCreateRequest(BaseModel):
    type: JobType
    process_type: ProcessType
    resource_id: UUID
    event_id: Optional[UUID] = None
    datastore_id: Optional[UUID] = None
    datastore_endpoint: Optional[str] = None
    state: ProcessState = ProcessState.ready


class JobUpdateRequest(BaseModel):
    type: JobType = optional_field
    process_type: ProcessType = optional_field
    resource_id: UUID = optional_field
    datastore_id: Optional[UUID] = optional_field
    datastore_endpoint: Optional[str] = optional_field
    state: ProcessState = optional_field
    error: Optional[dict] = optional_field



from typing import Dict, Type

from lqs.transcode.cutter.cutter import *
import lqs.transcode.cutter.cutter as CutterClasses
import orjson

_custom_classes: Dict[str, Type[CustomClass]] = {
    # "lqs_cutter/toro.nav.inputs.IMU": bhc_navcon_imu,
    # "lqs_cutter/navcon_imu": navcon_imu,
    "toro.nav.inputs.IMU": bhc_navcon_imu,
    "navcon.imu": navcon_imu,
    "navcon_internal.health.heartbeat": navcon_reporter_heartbeat,
    "navcon.0.pose_filter.cal_data": navcon_pose_filter_calibration_data,
    # 'navcon.0.pose_filter.binary':navcon_pose_filter,
    "navcon.stats.temperature": navcon_temperature,
    "navcon_internal.0.img.metadata": navcon_image_metadata,
    "platform.heartbeat": navcon_heartbeat,
    "platform.gps": navcon_gps,
    "navcon.heartbeat": navcon_heartbeat,
    "navcon.0.localizer.cam_select": navcon_localizer_cam_select,
    "navcon_internal.0.autoexposure_status": navcon_autoexposure_status,
    "navcon.0.platform_response.speed_limit": navcon_speed_limit,
    "navcon.0.platform_request.speed_limit": navcon_speed_limit,
    "navcon_internal.camera_pose_success": navcon_camera_pose_success,
    "navcon_internal.img.docking_called_shot.1": navcon_image,
    "navcon_internal.img.called_shot.1": navcon_image,
    "navcon_internal.dock.called_shot": navcon_image,
    "navcon_internal.img.docking_called_shot.2": navcon_image,
    "navcon_internal.img.called_shot.0": navcon_image,
    "navcon_internal.img.called_shot.2": navcon_image,
    "navcon_internal.img.docking_called_shot.0": navcon_image,
    "navcon_internal.img.called_shot.3": navcon_image,
    "navcon_internal.img.docking_called_shot.3": navcon_image,
    "navcon_internal.localization_img.1": navcon_image,
    "navcon_internal.localization_img.2": navcon_image,
    "navcon_internal.localization_img.3": navcon_image,
    "navcon_internal.localization_img.0": navcon_image,
    "navcon.0.localizer.init_loc_info": navcon_init_localization_info,
    "navcon_internal.0.localizer_config": navcon_localizer_config,
    "navcon.gps": navcon_gps,
    "navcon.health.info": navcon_health_info,
    "events.encoder_stillness": navcon_event,
    "navcon.0.localizer.initialization_results": navcon_localizer_initialization_results,
    "navcon.0.telem.fsm_state": navcon_telemetry_state_change,
    # json_classes
    "navcon.0.pose_filter.json": StringifiedJson,
    "navcon_internal.stats.temperature.json": StringifiedJson,
    "navcon_internal.stats.memory.json": StringifiedJson,
    "navcon_internal.stats.trip.json": StringifiedJson,
    "navcon_internal.stats.cooling.json": StringifiedJson,
    "navcon_internal.stats.cpu.json": StringifiedJson,
    "navcon_internal.0.web_heartbeat.json": StringifiedJson,
    "navcon_internal.imu_debug.json": StringifiedJson,
    "navcon_internal.imu_timing.json": StringifiedJson,
    # new topics from Steve Landers
    "platform.wheelencoders": navcon_encoder_update,
    "navcon.0.pose_filter.binary": navcon_pose_filter,
    "navcon.0.pose_filter.binary.throttled": navcon_pose_filter,
    "navcon.0.pose_filter.docking.json": StringifiedJson,
    "navcon.0.pose_filter.docking.binary": navcon_pose_filter,
    "debug.hardware_id": StringifiedJson,
    "navcon.state.desired": navcon_desired_state,
    "navcon_internal.health.fault": navcon_fault_message,
    "navcon.called_shot": navcon_called_shot,
    "debug.unit_info": StringifiedJson,
    "debug.unit_config": StringifiedJson,
    "debug.application_config": StringifiedJson,
    "navcon.health.system": navcon_health,
    "debug.software_version": StringifiedJson,
    "navcon.0.localizer.safety_poses": navcon_safety_localization_poses,
}

wild_card_prefixes = {
    "navcon_internal.img.called_shot.*": navcon_called_shot,
    "cutter.v4lcam.img.*": navcon_image,
    "navcon_internal.localization_img.*": navcon_image,
}


def get_type_name_for_cutter(topic_name):
    type_name = _custom_classes.get(topic_name, None)
    if type_name:
        type_name = f"lqs_cutter/{type_name.__name__}"
    else:
        for wildcard_entry, cls in wild_card_prefixes.items():
            prefix = wildcard_entry.split("*")[0]
            if topic_name.startswith(prefix):
                type_name = f"lqs_cutter/{cls.__name__}"
                break
    return type_name


class CutterDecoder:
    def deserialize(self, type_name: str, payload: bytes):
        assert type_name.startswith("lqs_cutter/")
        type_name = type_name.removeprefix("lqs_cutter/")
        custom_class = getattr(CutterClasses, type_name)
        res = custom_class.deserialize(payload)
        return orjson.loads(orjson.dumps(res))



from lqs.interface.dsm.models.announcement import *
from lqs.interface.dsm.models.api_key import *
from lqs.interface.dsm.models.comment import *
from lqs.interface.dsm.models.configuration import *
from lqs.interface.dsm.models.datastore import *
from lqs.interface.dsm.models.datastore_association import *
from lqs.interface.dsm.models.event import *
from lqs.interface.dsm.models.inference import *
from lqs.interface.dsm.models.job import *
from lqs.interface.dsm.models.manifest import *
from lqs.interface.dsm.models.role import *
from lqs.interface.dsm.models.ticket import *
from lqs.interface.dsm.models.usage_record import *
from lqs.interface.dsm.models.user import *
from lqs.interface.dsm.models.__common__ import *



import sys
import json
import traceback


class BadRequestException(Exception):
    def __init__(self, msg=None):
        self.message = (
            msg
            if msg is not None
            else "The request couldn't be processed due to a conflict."
        )
        super().__init__(self.message)


class UnauthorizedException(Exception):
    def __init__(self, msg=None):
        self.message = (
            msg
            if msg is not None
            else "The request couldn't be processed since the user is not authenticated."
        )
        super().__init__(self.message)


class ForbiddenException(Exception):
    def __init__(self, action=None, resource=None, reason=None, msg=None):
        if msg is not None:
            self.message = msg
        else:
            self.message = (
                f"Action '{action}' on '{resource}' not allowed for the current user."
            )
            if reason is not None:
                self.message = f"Action '{action}' on '{resource}' not allowed for the current user: {reason}"
        super().__init__(self.message)


class NotFoundException(Exception):
    def __init__(self, resource=None, resource_id=None, msg=None):
        if msg is not None:
            self.message = msg
        else:
            if resource is not None:
                if resource_id is not None:
                    self.message = f"{resource} with ID {resource_id} not found."
                else:
                    self.message = f"{resource} not found."
            else:
                self.message = "Resource not found."
        super().__init__(self.message)


class InternalServerErrorException(Exception):
    def __init__(self, msg=None):
        if msg is not None:
            self.message = msg
        else:
            (
                exception_type,
                exception_value,
                exception_traceback,
            ) = sys.exc_info()
            traceback_string = traceback.format_exception(
                exception_type, exception_value, exception_traceback
            )
            err_msg = {
                "errorType": exception_type.__name__
                if exception_type
                else "UnknownError",
                "errorMessage": str(exception_value),
                "stackTrace": traceback_string,
            }
            self.message = json.dumps(err_msg)
        super().__init__(self.message)


class ConflictException(Exception):
    def __init__(self, msg=None):
        self.message = (
            msg
            if msg is not None
            else "The request couldn't be processed due to a conflict."
        )
        super().__init__(self.message)


class ContentTooLargeException(Exception):
    def __init__(self, msg=None):
        self.message = (
            msg
            if msg is not None
            else "The request couldn't be processed due to the content being too large."
        )
        super().__init__(self.message)


class LockedException(Exception):
    def __init__(self, msg=None):
        self.message = (
            msg
            if msg is not None
            else "The request couldn't be processed due to a lock."
        )
        super().__init__(self.message)


class NotImplementedException(Exception):
    def __init__(self, msg=None):
        self.message = (
            msg
            if msg is not None
            else "The request couldn't be processed since the requested service is not implemented."
        )
        super().__init__(self.message)


class TimeoutException(Exception):
    def __init__(self, msg=None):
        self.message = (
            msg
            if msg is not None
            else "The request couldn't be processed due to a timeout."
        )
        super().__init__(self.message)


class DataStoreNotFoundException(Exception):
    def __init__(self, datastore_name=None):
        self.message = "DataStore not found."
        if datastore_name is not None:
            self.message = f"DataStore {datastore_name} not found."


class LogIntegrityException(Exception):
    def __init__(self, msg=None):
        self.message = (
            msg
            if msg is not None
            else "The request couldn't be processed due to a log integrity error."
        )
        super().__init__(self.message)


# Process Exceptions


class AsyncIngestionError(Exception):
    pass



from typing import Optional
from uuid import UUID

from pydantic import BaseModel

from lqs.interface.base.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)
from lqs.interface.base import models


class APIKey(CommonModel["APIKey"]):
    name: str
    user_id: UUID
    disabled: bool
    secret: Optional[str]

    def fetch_user(self) -> "models.User":
        return self._app.fetch.user(self.user_id).data


class APIKeyDataResponse(DataResponseModel[APIKey]):
    pass


class APIKeyListResponse(PaginationModel[APIKey]):
    pass


class APIKeyCreateRequest(BaseModel):
    name: str
    user_id: UUID
    disabled: bool = False


class APIKeyUpdateRequest(BaseModel):
    name: str = optional_field
    disabled: bool = optional_field



import orjson
import json
from dataclasses import asdict, is_dataclass
from rosbags.typesys import (
    Stores,
    get_types_from_msg,
    get_typestore,
)
import numpy as np


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif is_dataclass(obj):
            return asdict(obj)
        return json.JSONEncoder.default(self, obj)


class CDRWrapper:
    def __init__(self):
        self.type_store = get_typestore(Stores.LATEST)

    def deserialize(self, type_name, type_data, message_bytes):
        if type_name not in self.type_store.fielddefs:
            types_dict = get_types_from_msg(text=type_data, name=type_name)
            self.type_store.register(typs=types_dict)

        res = self.type_store.deserialize_cdr(
            rawdata=message_bytes,
            typename=type_name,
        )
        return orjson.loads(orjson.dumps(res, default=NumpyEncoder().default))



from lqs.transcode.ros1.RosMessageParser import RosMessageParser


from antlr4 import ParseTreeVisitor
import hashlib


def compute_md5(msg_text: str):
    md5 = hashlib.md5()
    md5.update(msg_text.encode("utf-8"))
    return md5.hexdigest()


def compute_md5_from_dict(schema: dict):
    # it has constants and properties
    msg_text = "\n".join([*schema["constants"], *schema["properties"]]).strip()
    return compute_md5(msg_text)


class RosMessageParserVisitorForMD5(ParseTreeVisitor):
    def __init__(self, message_class_name: str):
        self.root_class_name = message_class_name
        self.md5_registry = {}

    # Visit a parse tree produced by RosMessageParser#rosbag_input.
    def visitRosbag_input(self, ctx: RosMessageParser.Rosbag_inputContext):
        # visit children in reverse order, that makes it easier since the dependencies get processed first
        for child in reversed(ctx.children[1:]):
            self.visit(tree=child)
        self.message_class_name = self.root_class_name
        if "/" in self.message_class_name:
            self.current_package = self.message_class_name.split("/")[0]
        self.visit(ctx.children[0])
        return self.md5_registry[self.root_class_name]

    # Visit a parse tree produced by RosMessageParser#rosbag_nested_message.
    def visitRosbag_nested_message(
        self, ctx: RosMessageParser.Rosbag_nested_messageContext
    ):
        self.message_class_name = ctx.getChild(1).getText()
        if "/" in self.message_class_name:
            self.current_package = self.message_class_name.split("/")[0]
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#ros_message.
    def visitRos_message(self, ctx: RosMessageParser.Ros_messageContext):
        # collect all the non-comment children
        member_declaration_list = list(
            ctx.getChildren(
                lambda x: not isinstance(x, RosMessageParser.CommentContext)
            )
        )
        # split into constants and members
        constant_declaration_list = [
            member
            for member in member_declaration_list
            if isinstance(member, RosMessageParser.Constant_declarationContext)
        ]
        member_declaration_list = [
            member
            for member in member_declaration_list
            if isinstance(member, RosMessageParser.Field_declarationContext)
        ]

        schema = {
            "type": "object",
            "constants": [self.visit(member) for member in constant_declaration_list],
            "properties": [self.visit(member) for member in member_declaration_list],
        }
        self.md5_registry[self.message_class_name] = compute_md5_from_dict(schema)

    # Visit a parse tree produced by RosMessageParser#field_declaration.
    def visitField_declaration(self, ctx: RosMessageParser.Field_declarationContext):
        member_name = ctx.getChild(1).getText()
        member_type = self.visit(ctx.getChild(0))
        return f"{member_type} {member_name}"

    # Visit a parse tree produced by RosMessageParser#constant_declaration.
    def visitConstant_declaration(
        self, ctx: RosMessageParser.Constant_declarationContext
    ):
        type = ctx.getChild(0).getText()
        member_name = ctx.getChild(1).getText()
        const_value = self.visit(ctx.getChild(3))
        return f"{type} {member_name}={const_value}"

    # Visit a parse tree produced by RosMessageParser#comment.
    def visitComment(self, ctx: RosMessageParser.CommentContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#identifier.
    def visitIdentifier(self, ctx: RosMessageParser.IdentifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#type.
    def visitType(self, ctx: RosMessageParser.Type_Context):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#ros_type.
    def visitRos_type(self, ctx: RosMessageParser.Ros_typeContext):
        class_type = ctx.getText()
        if class_type == "Header":
            # prefix with package as std_msgs
            class_type = "std_msgs/Header"
        if "/" not in class_type:
            class_type = f"{self.current_package}/{class_type}"
        if class_type not in self.md5_registry:
            return ""
        return self.md5_registry[class_type]

    # Visit a parse tree produced by RosMessageParser#array_type.
    def visitArray_type(self, ctx: RosMessageParser.Array_typeContext):
        # TODO handle this with the broken logic that ROS uses for custom classes
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#variable_array_type.
    def visitVariable_array_type(
        self, ctx: RosMessageParser.Variable_array_typeContext
    ):
        base_type = ctx.getChild(0).getChild(0)
        if isinstance(base_type, RosMessageParser.Ros_typeContext):
            # the broken logic is only for Complex Types
            return self.visit(base_type)
        else:
            return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#fixed_array_type.
    def visitFixed_array_type(self, ctx: RosMessageParser.Fixed_array_typeContext):
        # array, type, and fixed length
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#integral_type.
    def visitIntegral_type(self, ctx: RosMessageParser.Integral_typeContext):
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#floating_point_type.
    def visitFloating_point_type(
        self, ctx: RosMessageParser.Floating_point_typeContext
    ):
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#temporal_type.
    def visitTemporal_type(self, ctx: RosMessageParser.Temporal_typeContext):
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#string_type.
    def visitString_type(self, ctx: RosMessageParser.String_typeContext):
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#boolean_type.
    def visitBoolean_type(self, ctx: RosMessageParser.Boolean_typeContext):
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#sign.
    def visitSign(self, ctx: RosMessageParser.SignContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#integral_value.
    def visitIntegral_value(self, ctx: RosMessageParser.Integral_valueContext):
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#floating_point_value.
    def visitFloating_point_value(
        self, ctx: RosMessageParser.Floating_point_valueContext
    ):
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#bool_value.
    def visitBool_value(self, ctx: RosMessageParser.Bool_valueContext):
        return ctx.getText()

    # Visit a parse tree produced by RosMessageParser#string_value.
    def visitString_value(self, ctx: RosMessageParser.String_valueContext):
        return ctx.getText()



from typing import Optional

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    CommonModel,
    ProcessType,
    DataResponseModel,
    PaginationModel,
    optional_field,
)
from lqs.interface.core import models


class Workflow(CommonModel):
    name: str = Field(..., description="The unique name of the workflow for reference.")
    process_type: Optional[ProcessType] = Field(
        None,
        description="The process type of the workflow (`ingestion` or `digestion`).",
    )
    note: Optional[str] = Field(
        None, description="A general note about the workflow for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the workflow."
    )
    managed: bool = Field(
        ..., description="Whether the workflow is managed by the system."
    )
    default: bool = Field(
        ..., description="Whether the workflow is the default workflow."
    )
    disabled: bool = Field(
        ..., description="Whether the workflow is disabled and cannot be used."
    )
    context_schema: Optional[dict] = Field(
        None,
        description="The context JSON schema for the workflow which process workflow contexts are validated against.",
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            [
                "process_type",
                "note",
                "context",
                "context_schema",
            ]
        )
    )

    def list_digestions(self, threaded=False, **kwargs) -> list["models.Digestion"]:
        return self._list_all_subresources(
            list_method=self._app.list.digestion,
            threaded=threaded,
            workflow_id=self.id,
            **kwargs,
        )

    def list_ingestions(self, threaded=False, **kwargs) -> list["models.Ingestion"]:
        return self._list_all_subresources(
            list_method=self._app.list.ingestion,
            threaded=threaded,
            workflow_id=self.id,
            **kwargs,
        )

    def list_hooks(self, threaded=False, **kwargs) -> list["models.Hook"]:
        return self._list_all_subresources(
            list_method=self._app.list.hook,
            threaded=threaded,
            workflow_id=self.id,
            **kwargs,
        )


class WorkflowDataResponse(DataResponseModel[Workflow]):
    pass


class WorkflowListResponse(PaginationModel[Workflow]):
    pass


class WorkflowCreateRequest(BaseModel):
    name: str
    process_type: ProcessType = ProcessType.digestion
    note: Optional[str] = None
    context: Optional[dict] = None
    default: bool = False
    disabled: bool = False
    managed: bool = False
    context_schema: Optional[dict] = None


class WorkflowUpdateRequest(BaseModel):
    name: str = optional_field
    note: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    default: bool = optional_field
    disabled: bool = optional_field
    context_schema: Optional[dict] = optional_field



from typing import Union, Optional

from lqs.common.facade import Facade

from lqs.client.config import RESTClientConfig
from lqs.client.core import LogQS
from lqs.client.dsm import DataStoreManager
from lqs.client.admin import Admin
from lqs.client.utils import Utils


class RESTClient(Facade):
    config: RESTClientConfig
    version = "1.1"

    def __init__(
        self,
        config: Union[RESTClientConfig, dict] = RESTClientConfig(),
        http_client=None,
        datastore_id: Optional[str] = None,
    ):
        self.setup(config, RESTClientConfig)
        self.http_client = http_client
        self.datastore_id = datastore_id

        self.lqs = LogQS(app=self)
        self.dsm = DataStoreManager(app=self)
        self.admin = Admin(app=self)

        self.list = self.lqs.list
        self.fetch = self.lqs.fetch
        self.create = self.lqs.create
        self.update = self.lqs.update
        self.delete = self.lqs.delete
        self.resource = self.lqs.resource

        self.utils = Utils(app=self)

    def set_datastore_id(self, datastore_id: Optional[str]):
        self.datastore_id = datastore_id
        return self

    def get_datastore_id(self):
        return self.datastore_id or self.config.datastore_id

    def set_http_client(self, http_client):
        self.http_client = http_client
        return self

    def get_http_client(self):
        return self.http_client



from typing import List, Optional
from uuid import UUID

from pydantic import Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    CommonModel,
    DataResponseModel,
    LockModel,
    LockCreateRequest,
    LockUpdateRequest,
    PaginationModel,
    optional_field,
    optional_deprecated_field,
    Int64,
)

from lqs.interface.core import models


class Log(CommonModel, LockModel):
    group_id: UUID = Field(
        ..., description="The ID of the group to which this log belongs."
    )
    name: str = Field(..., description="The name of the log (unique per group).")

    start_time: Optional[Int64] = Field(
        None, description="The timestamp of the first record of the log."
    )
    end_time: Optional[Int64] = Field(
        None, description="The timestamp of the last record of the log."
    )
    duration: Optional[Int64] = Field(
        None, description="The duration of the log in nanoseconds."
    )
    base_timestamp: Optional[Int64] = Field(
        None,
        description="The time, in nanoseconds, to be added to all timestamps in the log.",
    )
    record_size: int = Field(
        ..., description="The total size of all records in the log in bytes."
    )
    record_count: int = Field(
        ..., description="The total number of records in the log."
    )
    object_size: int = Field(
        0,
        description="The total size of all objects in the log in bytes.",
        deprecated="This field is no longer populated with a non-zero value.",
    )
    object_count: int = Field(
        0,
        description="The total number of objects in the log.",
        deprecated="This field is no longer populated with a non-zero value.",
    )

    note: Optional[str] = Field(
        None, description="A general note about the log for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the log."
    )
    default_workflow_id: Optional[UUID] = Field(
        None,
        description="The ID of the workflow to be executed during state transitions of associated processes.",
        deprecated="This field is no longer populated with a non-null value.",
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            [
                "start_time",
                "end_time",
                "duration",
                "base_timestamp",
                "object_size",
                "object_count",
                "note",
                "context",
                "default_workflow_id",
            ]
        )
    )

    def list_topics(self, threaded=False, **kwargs) -> List["models.Topic"]:
        return self._list_all_subresources(
            list_method=self._app.list.topic,
            threaded=threaded,
            log_id=self.id,
            **kwargs,
        )

    def list_ingestions(self, threaded=False, **kwargs) -> List["models.Ingestion"]:
        return self._list_all_subresources(
            list_method=self._app.list.ingestion,
            threaded=threaded,
            log_id=self.id,
            **kwargs,
        )

    def list_digestions(self, threaded=False, **kwargs) -> List["models.Digestion"]:
        return self._list_all_subresources(
            list_method=self._app.list.digestion,
            threaded=threaded,
            log_id=self.id,
            **kwargs,
        )

    def fetch_group(self) -> "models.Group":
        return self._app.fetch.group(self.group_id).data


Log.model_rebuild()


class LogDataResponse(DataResponseModel[Log]):
    pass


class LogListResponse(PaginationModel[Log]):
    pass


class LogCreateRequest(LockCreateRequest):
    group_id: UUID
    name: str
    note: Optional[str] = None
    base_timestamp: Optional[Int64] = None
    context: Optional[dict] = None
    default_workflow_id: Optional[UUID] = Field(
        None,
        deprecated=True,
    )


class LogUpdateRequest(LockUpdateRequest):
    group_id: UUID = optional_field
    name: str = optional_field
    note: Optional[str] = optional_field
    base_timestamp: Optional[Int64] = optional_field
    context: Optional[dict] = optional_field
    default_workflow_id: Optional[UUID] = optional_deprecated_field



from abc import ABC, abstractmethod
from uuid import UUID

import lqs.interface.base.models as models


class UpdateInterface(ABC):
    def _process_data(self, data):
        if not isinstance(data, dict):
            return data.model_dump(exclude_unset=True)
        return data

    @abstractmethod
    def _api_key(self, **kwargs) -> models.APIKeyDataResponse:
        pass

    def api_key(
        self, api_key_id: UUID, data: models.APIKeyUpdateRequest
    ) -> models.APIKeyDataResponse:
        return self._api_key(
            api_key_id=api_key_id,
            data=self._process_data(data),
        )

    @abstractmethod
    def _role(self, **kwargs) -> models.RoleDataResponse:
        pass

    def role(
        self, role_id: UUID, data: models.RoleUpdateRequest
    ) -> models.RoleDataResponse:
        return self._role(
            role_id=role_id,
            data=self._process_data(data),
        )

    @abstractmethod
    def _user(self, **kwargs) -> models.UserDataResponse:
        pass

    def user(
        self, user_id: UUID, data: models.UserUpdateRequest
    ) -> models.UserDataResponse:
        return self._user(user_id=user_id, data=self._process_data(data))



from typing import List, Optional
from uuid import UUID

from pydantic import Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    CommonModel,
    DataResponseModel,
    LockModel,
    LockCreateRequest,
    LockUpdateRequest,
    PaginationModel,
    optional_field,
    optional_deprecated_field,
)
from lqs.interface.core import models


class Group(CommonModel, LockModel):
    name: str = Field(..., description="The unique name of the group.")
    note: Optional[str] = Field(
        None, description="A general note about the group for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the group."
    )
    default_workflow_id: Optional[UUID] = Field(
        None,
        description="The ID of the workflow to be executed during state transitions of associated processes.",
        deprecated="This field is no longer populated with a non-null value.",
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            ["note", "context", "default_workflow_id"]
        )
    )

    def list_logs(self, threaded=False, **kwargs) -> List["models.Log"]:
        return self._list_all_subresources(
            list_method=self._app.list.log,
            threaded=threaded,
            group_id=self.id,
            **kwargs,
        )


class GroupDataResponse(DataResponseModel[Group]):
    pass


class GroupListResponse(PaginationModel[Group]):
    pass


class GroupCreateRequest(LockCreateRequest):
    name: str
    note: Optional[str] = None
    context: Optional[dict] = None
    default_workflow_id: Optional[UUID] = Field(
        None,
        deprecated=True,
    )


class GroupUpdateRequest(LockUpdateRequest):
    name: str = optional_field
    note: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    default_workflow_id: Optional[UUID] = optional_deprecated_field



from uuid import UUID
from typing import Optional

from pydantic import BaseModel

from lqs.interface.dsm.models.__common__ import (
    DataResponseModel,
    TimeSeriesModel,
    PaginationModel,
    Int64,
)


class UsageRecord(TimeSeriesModel):
    _repr_fields = ("timestamp", "datastore_id")

    datastore_id: Optional[UUID]
    user_id: Optional[UUID]
    duration: Optional[Int64]
    category: Optional[str]
    usage_data: Optional[dict | list]


class UsageRecordDataResponse(DataResponseModel[UsageRecord]):
    pass


class UsageRecordListResponse(PaginationModel[UsageRecord]):
    pass


class UsageRecordCreateRequest(BaseModel):
    timestamp: Int64
    datastore_id: Optional[UUID]
    user_id: Optional[UUID]
    duration: Optional[Int64]
    category: Optional[str]
    usage_data: Optional[dict | list]



from typing import Optional
from uuid import UUID

from pydantic import Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    DataResponseModel,
    PaginationModel,
    ProcessModel,
    ProcessCreateRequest,
    ProcessUpdateRequest,
    optional_field,
    process_required_fields,
)
from lqs.interface.core import models


class Ingestion(ProcessModel["Ingestion"]):
    log_id: UUID = Field(
        ..., description="The ID of the log to which this ingestion belongs."
    )

    object_store_id: Optional[UUID] = Field(
        None,
        description="The ID of the object store where the ingestion object is stored.",
    )
    object_key: Optional[str] = Field(
        None, description="The key of the ingestion object."
    )

    name: Optional[str] = Field(
        None, description="The name of the ingestion for reference (not unique)."
    )
    note: Optional[str] = Field(
        None, description="A general note about the ingestion for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the ingestion."
    )

    group_id: UUID = Field(
        ...,
        description="The ID of the group to which this ingestion belongs.",
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            ["object_store_id", "object_key", "name", "note", "context"]
            + process_required_fields
        )
    )

    def list_ingestion_parts(
        self, threaded=False, **kwargs
    ) -> list["models.IngestionPart"]:
        return self._list_all_subresources(
            list_method=self._app.list.ingestion_part,
            threaded=threaded,
            ingestion_id=self.id,
            **kwargs,
        )

    def fetch_log(self) -> "models.Log":
        return self._app.fetch.log(self.log_id).data

    def fetch_group(self) -> "models.Group":
        return self._app.fetch.group(self.group_id).data

    def fetch_workflow(self) -> Optional["models.Workflow"]:
        if self.workflow_id is None:
            return None
        return self._app.fetch.workflow(self.workflow_id).data


class IngestionDataResponse(DataResponseModel[Ingestion]):
    pass


class IngestionListResponse(PaginationModel[Ingestion]):
    pass


class IngestionCreateRequest(ProcessCreateRequest):
    log_id: UUID
    name: Optional[str] = None
    object_store_id: Optional[UUID] = None
    object_key: Optional[str] = None
    note: Optional[str] = None
    context: Optional[dict] = None


class IngestionUpdateRequest(ProcessUpdateRequest):
    name: Optional[str] = optional_field
    object_store_id: Optional[UUID] = optional_field
    object_key: Optional[str] = optional_field
    note: Optional[str] = optional_field
    context: Optional[dict] = optional_field
